# -* coding: utf-8 -*-

#################################################################################
# This file is part of EVALLIES.                                                #
#                                                                               #
# EVALLIES is a python package for lifelong learning speaker diarization.       #
# Home page: https://git-lium.univ-lemans.fr/Larcher/evallies                   #
#                                                                               #
# EVALLIES is free software: you can redistribute it and/or modify              #
# it under the terms of the GNU LLesser General Public License as               #
# published by the Free Software Foundation, either version 3 of the License,   #
# or (at your option) any later version.                                        #
#                                                                               #
# EVALLIES is distributed in the hope that it will be useful,                   #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 #
# GNU Lesser General Public License for more details.                           #
#                                                                               #
# You should have received a copy of the GNU Lesser General Public License      #
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.              #
#                                                                               #
#################################################################################
"""
Copyright 2020-2021 Anthony Larcher, Meysam Shamsi & Yevhenii Propkopalo

    :mod:`lium_baseline.interactive`

"""

import copy
import logging
import numpy
import pandas
import random
import scipy
import os

from s4d.clustering.hac_utils import scores2distance
from scipy.cluster import hierarchy as hac
from scipy.spatial.distance import squareform

from .utils import s4d_to_allies
from ..user_simulation import MessageToUser
from ..user_simulation import Request
from s4d.clustering.hac_iv import information
from ..der_single import *

from s4d.diar import Diar
from s4d.scoring import DER
import json

def allies_write_diar(current_diar, filename):
    """

    """
    cdiar = copy.deepcopy(current_diar)
    for idx, seg in enumerate(cdiar):
        cdiar.segments[idx]['start'] = float(seg['start']) / 100.
        cdiar.segments[idx]['stop'] = float(seg['stop']) / 100.


    cdiar.sort(['show', 'start'])
    with open(filename, 'w', encoding="utf8") as fic:
        for line in Diar.to_string_seg(cdiar, time_float=True):
            fic.write(line)


def get_cluster_nodes(scores,node_to_check,link):
    
    number_cluster=len(scores.modelset)
    node_list1=get_node_spkeakers(node_to_check[0],number_cluster,link)
    node_list2=get_node_spkeakers(node_to_check[1],number_cluster,link)
    #convert node_list to spk_list

    spk_list1=[]
    for n in node_list1:
        spk_list1.append(scores.modelset[n])
    spk_list2=[]
    for n in node_list2:
        spk_list2.append(scores.modelset[n])
        
        
        
def get_node_spkeakers(node_spk, number_cluster, link):
    """
    get a node and return the all sub cluster id

    :param node_spk: row of the linkage matrix
    :param number_cluster: number in link matrix
    :param link: linkage matrix
    :return: cluster_list: a list of clusters id in link
    """

    if node_spk >= number_cluster:
        cluster_list = get_node_spkeakers(link[int(node_spk-number_cluster)][0],
                                          number_cluster,
                                          link) + get_node_spkeakers(link[int(node_spk - number_cluster)][1],
                                                                     number_cluster,
                                                                     link)
    else:
        cluster_list=[int(node_spk)]

    return cluster_list


def sort_node_sides_ccs(current_diar,
                        first_spk,
                        second_spk,
                        scores_per_segment,
                        condidate_pair_number=None):
    """
    provide two lists of segments by which has maximum plda similarity scores to cluster's center
    :param current_diar:
    :param current_vec:
    :return:
    """
    #finding the index of segment in current_diar for investigation in scores_per_segment
    segs_indx_cluster1=[]
    for i in range(len(current_diar)):
        if current_diar[i][1] in first_spk:
            segs_indx_cluster1.append(i)
    segs_indx_cluster2=[]
    for i in range(len(current_diar)):
        if current_diar[i][1] in second_spk:
            segs_indx_cluster2.append(i)

    similarity_scoremat=copy.deepcopy(scores_per_segment.scoremat)
    
    
    #calculate the sum of similarity to cluster's segments
    cluster1_similaritySum={}
    for i in range(len(segs_indx_cluster1)):
        for j in range(len(scores_per_segment.scoremat[i])):
            if j not in cluster1_similaritySum:
                cluster1_similaritySum[j]=0
            if j in segs_indx_cluster1 and j!=i:
                 cluster1_similaritySum[j]+=scores_per_segment.scoremat[i][j]
    cluster2_similaritySum={}
    for i in range(len(segs_indx_cluster2)):
        for j in range(len(scores_per_segment.scoremat[i])):
            if j not in cluster2_similaritySum:
                cluster2_similaritySum[j]=0
            if j in segs_indx_cluster2 and j!=i:
                 cluster2_similaritySum[j]+=scores_per_segment.scoremat[i][j]
                    
    #creat a list of combinition pair and their correspond additional similarity from cluster center as score
    list_pair=[]
    list_pair_scores=[]
    for s1 in cluster1_similaritySum:
        for s2 in cluster2_similaritySum:
            list_pair.append([s1,s2])
            list_pair_scores.append(cluster1_similaritySum[s1]+cluster2_similaritySum[s2])
            
    #sort the list_pair based on list_pair_scores
    _, list_pair_sorted = (list(t) for t in zip(*sorted(zip(list_pair_scores,
                                                             list_pair))))                
    #we are looking for maximum the similarity score
    list_pair_sorted.reverse()
    
    first_seg_list_sorted=[]
    second_seg_list_sorted=[]
    for p in list_pair_sorted:
        first_seg_list_sorted.append(current_diar[p[0]])
        second_seg_list_sorted.append(current_diar[p[1]])
    
    if condidate_pair_number==None:
        return first_seg_list_sorted, second_seg_list_sorted
    else:
        return first_seg_list_sorted[:condidate_pair_number], second_seg_list_sorted[:condidate_pair_number]


def sort_node_sides_cc(current_diar,
                        first_spk,
                        second_spk,
                        current_vec,
                        condidate_pair_number=None):
    """
    provide two lists of segments by which has minimum sum of distances from cluster centers
    :param current_diar:
    :param current_vec:
    :return:
    """
    #finding the index and correspond vectore of segment in current_diar for each cluster
    segs_indx2vec_cluster1={}
    for i in range(len(current_diar)):
        if current_diar[i][1] in first_spk:
            segs_indx2vec_cluster1[i]=current_vec.stat1[i]
    segs_indx2vec_cluster2={}
    for i in range(len(current_diar)):
        if current_diar[i][1] in second_spk:
            segs_indx2vec_cluster2[i]=current_vec.stat1[i]
    
    #calculate cluster center of each cluster
    cluster1_center=numpy.mean(list(segs_indx2vec_cluster1.values()), axis=0)
    cluster2_center=numpy.mean(list(segs_indx2vec_cluster2.values()), axis=0) 
    
    #calculate the distance between each segment from cluster center
    segs_indx2distanceCC_cluster1={}
    for s in segs_indx2vec_cluster1:
        segs_indx2distanceCC_cluster1[s]=scipy.spatial.distance.euclidean(segs_indx2vec_cluster1[s],
                                                                          cluster1_center)
    segs_indx2distanceCC_cluster2={}
    for s in segs_indx2vec_cluster2:
        segs_indx2distanceCC_cluster2[s]=scipy.spatial.distance.euclidean(segs_indx2vec_cluster2[s],
                                                                          cluster2_center)
        
    #creat a list of combinition pair and their correspond additional distance from cluster center as score
    list_pair=[]
    list_pair_scores=[]
    for s1 in segs_indx2distanceCC_cluster1:
        for s2 in segs_indx2distanceCC_cluster2:
            list_pair.append([s1,s2])
            list_pair_scores.append(segs_indx2distanceCC_cluster1[s1]+segs_indx2distanceCC_cluster2[s2])
    
    #sort the list_pair based on list_pair_scores
    _, list_pair_sorted = (list(t) for t in zip(*sorted(zip(list_pair_scores,
                                                             list_pair))))                
    
    first_seg_list_sorted=[]
    second_seg_list_sorted=[]
    for p in list_pair_sorted:
        first_seg_list_sorted.append(current_diar[p[0]])
        second_seg_list_sorted.append(current_diar[p[1]])
    
    if condidate_pair_number==None:
        return first_seg_list_sorted, second_seg_list_sorted
    else:
        return first_seg_list_sorted[:condidate_pair_number], second_seg_list_sorted[:condidate_pair_number]


def sort_node_sides_min(current_diar,
                        first_spk,
                        second_spk,
                        scores_per_segment,
                        condidate_pair_number=None):
    """
    provide two lists of segments by which has minimum distance in scores_per_segment

    :param first_seg_list:
    :param second_seg_list:
    :param scores_per_segment:
    :param current_diar:
    :param selection_criteria:
    :return:
    """
 
    #finding the index of segment in current_diar for investigation in scores_per_segment
    segs_indx_cluster1=[]
    for i in range(len(current_diar)):
        if current_diar[i][1] in first_spk:
            segs_indx_cluster1.append(i)
    segs_indx_cluster2=[]
    for i in range(len(current_diar)):
        if current_diar[i][1] in second_spk:
            segs_indx_cluster2.append(i)

    similarity_scoremat=copy.deepcopy(scores_per_segment.scoremat)

    first_seg_list_sorted=[]
    second_seg_list_sorted=[]
    
    #finding the max length of first_seg_list_sorted and second_seg_list_sorted
    if condidate_pair_number==None:
        _seg_list_sorted_len=len(segs_indx_cluster1)*len(segs_indx_cluster2)
    else:
        _seg_list_sorted_len=min(len(segs_indx_cluster1)*len(segs_indx_cluster2),condidate_pair_number)
    
    while (len(first_seg_list_sorted)<_seg_list_sorted_len and
           numpy.any((similarity_scoremat != -float('inf')))):
        
        #finding the index of two segments with minimum distance or maximum similarity
        first_seg_idx,second_seg_idx=numpy.unravel_index(similarity_scoremat.argmax(),
                                                         similarity_scoremat.shape)
        if (first_seg_idx in segs_indx_cluster1 and second_seg_idx in segs_indx_cluster2):
            first_seg_list_sorted.append(current_diar[first_seg_idx])
            second_seg_list_sorted.append(current_diar[second_seg_idx])
        similarity_scoremat[first_seg_idx,second_seg_idx]=-float('inf')

        
    return first_seg_list_sorted, second_seg_list_sorted
                
                
def sort_node_sides_max(current_diar,
                        first_spk,
                        second_spk,
                        scores_per_segment,
                        condidate_pair_number=None):
    """
    provide two lists of segments by which has maximum distance in scores_per_segment

    :param first_seg_list:
    :param second_seg_list:
    :param scores_per_segment:
    :param current_diar:
    :param current_vec:
    :param selection_criteria:
    :return:
    """
    
    #finding the index of segment in current_diar for investigation in scores_per_segment
    segs_indx_cluster1=[]
    for i in range(len(current_diar)):
        if current_diar[i][1] in first_spk:
            segs_indx_cluster1.append(i)
    segs_indx_cluster2=[]
    for i in range(len(current_diar)):
        if current_diar[i][1] in second_spk:
            segs_indx_cluster2.append(i)

    similarity_scoremat=copy.deepcopy(scores_per_segment.scoremat)

    first_seg_list_sorted=[]
    second_seg_list_sorted=[]
    
    #finding the max length of first_seg_list_sorted and second_seg_list_sorted
    if condidate_pair_number==None:
        _seg_list_sorted_len=len(segs_indx_cluster1)*len(segs_indx_cluster2)
    else:
        _seg_list_sorted_len=min(len(segs_indx_cluster1)*len(segs_indx_cluster2),condidate_pair_number)
    
    while (len(first_seg_list_sorted)<_seg_list_sorted_len and
           numpy.any((similarity_scoremat != float('inf')))):
        
        #finding the index of two segments with maximum distance or minimum similarity
        first_seg_idx,second_seg_idx=numpy.unravel_index(similarity_scoremat.argmin(),
                                                         similarity_scoremat.shape)
        if (first_seg_idx in segs_indx_cluster1 and second_seg_idx in segs_indx_cluster2):
            first_seg_list_sorted.append(current_diar[first_seg_idx])
            second_seg_list_sorted.append(current_diar[second_seg_idx])
        similarity_scoremat[first_seg_idx,second_seg_idx]=float('inf')

 
    return first_seg_list_sorted, second_seg_list_sorted


def sort_node_sides_random(first_seg_list,
                           second_seg_list,
                           condidate_pair_number=None):
    """
    Shuffle the two lists

    :param first_seg_list:
    :param second_seg_list:
    :param condidate_pair_number: number of condidate pair as a list. None for maximum length
    :return:
    """
    first_seg_list_ = copy.deepcopy(first_seg_list)
    second_seg_list_ = copy.deepcopy(second_seg_list)
    
    tmp_list=[]
    for seg1 in first_seg_list_:
        for seg2 in second_seg_list_:
            tmp_list.append([seg1,seg2])
    random.shuffle(tmp_list)
    first_seg_list_sorted=[]
    second_seg_list_sorted=[]
    for seg_pair in tmp_list:
        first_seg_list_sorted.append(seg_pair[0])
        second_seg_list_sorted.append(seg_pair[1])

    if condidate_pair_number==None:
        return first_seg_list_sorted, second_seg_list_sorted
    else:
        return first_seg_list_sorted[:condidate_pair_number], second_seg_list_sorted[:condidate_pair_number]


def sort_node_sides_longest(first_seg_list,
                            second_seg_list,
                            condidate_pair_number=None):
    """
    Sort each list of segments by descending duration

    :param first_seg_list:
    :param second_seg_list:
    :param condidate_pair_number: number of condidate pair as a list. None for maximum length
    :return:
    """
    # For each segment list, sort according to the length of the segments
    first_seg_list_sorted = []
    second_seg_list_sorted = []

    first_duration = numpy.zeros(len(first_seg_list))
    for idx, seg in enumerate(first_seg_list):
        first_duration[idx] = seg["stop"] - seg["start"]

    first_indices = numpy.argsort(first_duration)
    for ii in first_indices[::-1]:
        first_seg_list_sorted.append(first_seg_list[ii])

    second_duration = numpy.zeros(len(second_seg_list))
    for idx, seg in enumerate(second_seg_list):
        second_duration[idx] = seg["stop"] - seg["start"]

    second_indices = numpy.argsort(second_duration)
    for ii in second_indices[::-1]:
        second_seg_list_sorted.append(second_seg_list[ii])

    if condidate_pair_number==None:
        return first_seg_list_sorted, second_seg_list_sorted
    else:
        return first_seg_list_sorted[:condidate_pair_number], second_seg_list_sorted[:condidate_pair_number]


def sort_node_sides(first_seg_list,
                    second_seg_list,
                    first_spk,
                    second_spk,
                    scores_per_cluster,
                    scores_per_segment,
                    current_diar,
                    current_vec,
                    selection_criteria,
                    condidate_pair_number=None):
    """
    Sort both lists according to the chosen criteria
    return two lists but sorted according to the chosen criteria

    :param first_seg_list:
    :param second_seg_list:
    :param scores:
    :param current_diar:
    :param selection_criteria: longest, cluster_center, cluster_center_plda, random, max_noBICHAC
    max, min, ideal
    :return:
    """
    if selection_criteria == "longest":
        first_seg_list_sorted, second_seg_list_sorted = sort_node_sides_longest(first_seg_list,
                                                                                second_seg_list,
                                                                                condidate_pair_number)

    elif selection_criteria == "cluster_center":
        first_seg_list_sorted, second_seg_list_sorted = sort_node_sides_cc(current_diar,
                                                                           first_spk,
                                                                           second_spk,
                                                                           current_vec,
                                                                           condidate_pair_number)

    elif selection_criteria == "cluster_center_score":
        first_seg_list_sorted, second_seg_list_sorted = sort_node_sides_ccs(current_diar,
                                                                            first_spk,
                                                                            second_spk,
                                                                            scores_per_segment,
                                                                            condidate_pair_number)

    elif selection_criteria == "max_noBICHAC":
        #first_seg_list_sorted, second_seg_list_sorted = sort_node_sides_max_NBH(first_seg_list,
        #                                                                        second_seg_list,
        #                                                                        scores_per_cluster,
        #                                                                        current_diar,
        #                                                                        current_vec,
        #                                                                        selection_criteria)
        print("Not implemented yet")
        pass

    elif selection_criteria == "max":
        first_seg_list_sorted, second_seg_list_sorted = sort_node_sides_max(current_diar,
                                                                           first_spk,
                                                                           second_spk,
                                                                           scores_per_segment,
                                                                           condidate_pair_number)
        
    elif selection_criteria == "min":
        first_seg_list_sorted, second_seg_list_sorted = sort_node_sides_min(current_diar,
                                                                           first_spk,
                                                                           second_spk,
                                                                           scores_per_segment,
                                                                           condidate_pair_number)

#     elif selection_criteria == "ideal":
        #first_seg_list_sorted, second_seg_list_sorted = sort_node_sides_ideal(first_seg_list,
        #                                                                      second_seg_list,
        #                                                                      scores_per_cluster,
        #                                                                      current_diar,
        #                                                                      current_vec,
        #                                                                      selection_criteria)
#         print("Not implemented yet")
#         pass

    elif selection_criteria == "random":
        first_seg_list_sorted, second_seg_list_sorted = sort_node_sides_random(first_seg_list,
                                                                               second_seg_list,
                                                                               condidate_pair_number=None)

    return first_seg_list_sorted, second_seg_list_sorted


def get_ideal_segments(node_to_check,
                        linkage_matrix,
                        scores_per_cluster,
                        current_diar,
                        ref_file):
    
    # Get the list of speakers for each side of the node
    first_clusters = get_node_spkeakers(node_to_check[0], scores_per_cluster.modelset.shape[0], linkage_matrix)
    first_spk = [scores_per_cluster.modelset[n] for n in first_clusters]

    second_clusters = get_node_spkeakers(node_to_check[1], scores_per_cluster.modelset.shape[0], linkage_matrix)
    second_spk = [scores_per_cluster.modelset[n] for n in second_clusters]
    
    #finding the index of segment in current_diar for investigation in scores_per_segment
    segs_indx_cluster1=[]
    for i in range(len(current_diar)):
        if current_diar[i][1] in first_spk:
            segs_indx_cluster1.append(i)
    segs_indx_cluster2=[]
    for i in range(len(current_diar)):
        if current_diar[i][1] in second_spk:
            segs_indx_cluster2.append(i)

    ref = ref_file.to_diar(current_diar[0]["show"])

    if len(ref)!=len(current_diar):
        print ("Error ! number of segment in diar are not same ..."+current_diar[0]['show'])
        import pdb
        pdb.set_trace()
    dict1 = {}      
    for indx in segs_indx_cluster1:
        if len(ref)>indx:
            if ref[indx]['cluster'] in dict1.keys():
                dict1[ref[indx]['cluster']]+=(ref[indx]['stop']-ref[indx]['start'])
            else:
                dict1[ref[indx]['cluster']]=(ref[indx]['stop']-ref[indx]['start'])

    dict2 = {}
    for indx in segs_indx_cluster2:
        if len(ref)>indx:
            if ref[indx]['cluster'] in dict2.keys():
                dict2[ref[indx]['cluster']]+=(ref[indx]['stop']-ref[indx]['start'])
            else:
                dict2[ref[indx]['cluster']]=(ref[indx]['stop']-ref[indx]['start'])

    if max(dict2, key = dict2.get) == max(dict1, key = dict1.get):
        return True , 0
    else:
        return False , 0


def get_segment_sorted_list(node_to_check,
                            linkage_matrix,
                            scores_per_cluster,
                            scores_per_segment,
                            current_diar,
                            current_vec,
                            selection_criteria="longest"):
    """

    :param node_to_check:
    :param linkage_matrix:
    :param scores_per_cluster:
    :param scores_per_segment:
    :param current_diar:
    :param current_vec:
    :param user:
    :param file_info:
    :param selection_criteria:
    :return:
    """

    # Get the list of speakers for each side of the node
    first_clusters = get_node_spkeakers(node_to_check[0], scores_per_cluster.modelset.shape[0], linkage_matrix)
    first_spk = [scores_per_cluster.modelset[n] for n in first_clusters]

    second_clusters = get_node_spkeakers(node_to_check[1], scores_per_cluster.modelset.shape[0], linkage_matrix)
    second_spk = [scores_per_cluster.modelset[n] for n in second_clusters]

    # Get the list of segments corresponding to each speaker
    first_seg_list = [seg for seg in current_diar.segments if seg["cluster"] in first_spk]
    second_seg_list = [seg for seg in current_diar.segments if seg["cluster"] in second_spk]

    # Sort the two lists of segments according to the chosen criteria
    first_seg_list_sorted, second_seg_list_sorted = sort_node_sides(first_seg_list,
                                                                    second_seg_list,
                                                                    first_spk,
                                                                    second_spk,
                                                                    scores_per_cluster,
                                                                    scores_per_segment,
                                                                    current_diar,
                                                                    current_vec,
                                                                    selection_criteria)

    return first_seg_list_sorted, second_seg_list_sorted


def ask_question_al(node_to_check,
                    linkage_matrix,
                    scores_per_cluster,
                    scores_per_segment,
                    init_diar,
                    update_diar,
                    current_vec,
                    user,
                    file_info,
                    ref_dir,
                    selection_criteria="longest"):
    """

    :param node_to_check:
    :param scores_per_cluster:
    :param scores_init:
    :param current_diar:
    :param current_vec:
    :param user:
    :param file_info:
    :param ref_dir:
    :param link:
    :param selection_criteria:
    :return:
    """
    
    
    if selection_criteria == "ideal":
        answer, hal = get_ideal_segments(node_to_check,
                                         linkage_matrix,
                                         scores_per_cluster,
                                         init_diar,
                                         ref_dir)
        return answer, hal
        
    else:
        first_seg_list_sorted, second_seg_list_sorted = get_segment_sorted_list(node_to_check,
                                                                                linkage_matrix,
                                                                                scores_per_cluster,
                                                                                scores_per_segment,
                                                                                init_diar,
                                                                                current_vec,
                                                                                selection_criteria)

   
        ## TODO: The segment should be selected from list or can be provided by expert
        ##### For now it will select the head of list
        selected_seg_first=first_seg_list_sorted[0]
        selected_seg_second=second_seg_list_sorted[0]

        # For first segments of each side, get the center of the segment in seconds
        t1 = (selected_seg_first["start"] + selected_seg_first["stop"]) / 200.
        t2 = (selected_seg_second["start"] + selected_seg_second["stop"]) / 200.

        # Ask the question to the user
        message_to_user = MessageToUser(file_info,
                                        s4d_to_allies(init_diar),
                                        Request('same', t1, t2))

        hal, answer = user.validate(message_to_user)

        # Return the answer from the user
        return answer.answer, hal


# def ask_question(node_to_check, scores,scores_init, current_diar,current_vec, user, file_info,link,selection_criteria="longest"):
#     """


#     :param node_to_check: row of the linkage matrix to check with the user
#     :param scores: the corresponding score object
#     :param current_diar: the current diarization
#     :param user: the user simulation
#     :param link: linkage matrix
#     :return: is_same_speaker: a boolean, True if the clustering at this node is approved by the user
#     """
 
#     # get one segment for each side of the node (if it's a leaf, it's easy, if qe already have a cluster we need to
#     # list all segments in this cluster and then find the most central segment in the cluster from the score matrix
#     # HERE WE NEED TO FIND TWO indices in the scores.scoremat matrix
#     number_cluster=len(scores.modelset)
#     node_list1=get_node_spkeakers(node_to_check[0],number_cluster,link)
#     node_list2=get_node_spkeakers(node_to_check[1],number_cluster,link)
#     #convert node_list to spk_list

#     spk_list1=[]
#     for n in node_list1:
#         spk_list1.append(scores.modelset[n])
#     spk_list2=[]
#     for n in node_list2:
#         spk_list2.append(scores.modelset[n])
    
#     # From the indices, get the borders of the segments in the Diar object
#     seg1=[0,0]
#     seg2=[0,0]
    
#     if selection_criteria == "longest":
#         for i in range(len(current_diar)):
#             if current_diar[i][1] in spk_list1:
#                 if current_diar[i][4]-current_diar[i][3]>seg1[1]-seg1[0]:
#                     seg1[1]=current_diar[i][4]
#                     seg1[0]=current_diar[i][3]
#             if current_diar[i][1] in spk_list2:
#                 if current_diar[i][4]-current_diar[i][3]>seg2[1]-seg2[0]:
#                     seg2[1]=current_diar[i][4]
#                     seg2[0]=current_diar[i][3]
    
#     elif selection_criteria == "cluster_center":
#         segs_cluster1={}
#         for i in range(len(current_diar)):
#             if current_diar[i][1] in spk_list1:
#                 segs_cluster1[i]=current_vec.stat1[i]
#         cluster1_center=numpy.mean(list(segs_cluster1.values()), axis=0)
#         min_dis=float('inf')
#         for seg in segs_cluster1:
#             dis=scipy.spatial.distance.euclidean(segs_cluster1[seg],cluster1_center)
#             if dis<min_dis:
#                 min_dis=dis
#                 seg1=[current_diar[int(seg)][3],current_diar[int(seg)][4]]
#         segs_cluster2={}
#         for i in range(len(current_diar)):
#             if current_diar[i][1] in spk_list2:
#                 segs_cluster2[i]=current_vec.stat1[i]
#         cluster2_center=numpy.mean(list(segs_cluster2.values()), axis=0)
#         min_dis=float('inf')
#         for seg in segs_cluster2:
#             dis=scipy.spatial.distance.euclidean(segs_cluster2[seg],cluster2_center)
#             if dis<min_dis:
#                 min_dis=dis
#                 seg2=[current_diar[int(seg)][3],current_diar[int(seg)][4]]
                
#     elif selection_criteria == "cluster_center_plda":
        
#         segs_cluster1=[]
#         for i in range(len(current_diar)):
#             if current_diar[i][1] in spk_list1:
#                 segs_cluster1.append(i)
#         cluster1_similaritySum=[]
#         for i in range(len(segs_cluster1)):
#             cluster1_similaritySum.append(0)
#             for j in range(len(scores_init.scoremat[i])):
#                 if j in segs_cluster1 and j!=i:
#                      cluster1_similaritySum[i]+=scores_init.scoremat[i][j]
                        
#         center1_seg=segs_cluster1[cluster1_similaritySum.index(max(cluster1_similaritySum))]
#         seg1=[current_diar[int(center1_seg)][3],current_diar[int(center1_seg)][4]]        
        
#         segs_cluster2=[]
#         for i in range(len(current_diar)):
#             if current_diar[i][1] in spk_list2:
#                 segs_cluster2.append(i)
#         cluster2_similaritySum=[]
#         for i in range(len(segs_cluster2)):
#             cluster2_similaritySum.append(0)
#             for j in range(len(scores_init.scoremat[i])):
#                 if j in segs_cluster2 and j!=i:
#                      cluster2_similaritySum[i]+=scores_init.scoremat[i][j]
                        
#         center2_seg=segs_cluster2[cluster2_similaritySum.index(max(cluster2_similaritySum))]
#         seg2=[current_diar[int(center2_seg)][3],current_diar[int(center2_seg)][4]]
        
#     elif selection_criteria == "random":
#         import random
#         segs_cluster1=[]
#         for i in range(len(current_diar)):
#             if current_diar[i][1] in spk_list1:
#                 segs_cluster1.append(i)
#         random_seg=random.choice(segs_cluster1)
#         seg1=[current_diar[int(random_seg)][3],current_diar[int(random_seg)][4]]
#         segs_cluster2=[]
#         for i in range(len(current_diar)):
#             if current_diar[i][1] in spk_list2:
#                 segs_cluster2.append(i)
#         random_seg=random.choice(segs_cluster2)
#         seg2=[current_diar[int(random_seg)][3],current_diar[int(random_seg)][4]]


        
#     elif selection_criteria == "max":
        
#         segs_cluster1=[]
#         for i in range(len(current_diar)):
#             if current_diar[i][1] in spk_list1:
#                 segs_cluster1.append(i)
        
#         segs_cluster2=[]
#         for i in range(len(current_diar)):
#             if current_diar[i][1] in spk_list2:
#                 segs_cluster2.append(i)
                  
#         min_similardis=float('inf')
#         for i in segs_cluster1:
#             for j in segs_cluster2:
#                 if scores_init.scoremat[i][j]< min_similardis:
#                     min_similardis=scores_init.scoremat[i][j]
#                     seg1=[current_diar[i][3],current_diar[i][4]]
#                     seg2=[current_diar[j][3],current_diar[j][4]]
                             
        
#     elif selection_criteria == "min":
#         segs_cluster1=[]
#         for i in range(len(current_diar)):
#             if current_diar[i][1] in spk_list1:
#                 segs_cluster1.append(i)
        
#         segs_cluster2=[]
#         for i in range(len(current_diar)):
#             if current_diar[i][1] in spk_list2:
#                 segs_cluster2.append(i)
                  
#         max_similardis=-float('inf')
#         for i in segs_cluster1:
#             for j in segs_cluster2:
#                 if scores_init.scoremat[i][j]> max_similardis:
#                     max_similardis=scores_init.scoremat[i][j]
#                     seg1=[current_diar[i][3],current_diar[i][4]]
#                     seg2=[current_diar[j][3],current_diar[j][4]]

#     elif selection_criteria == "max_noBICHAC":

#         segs_cluster1=node_list1
#         segs_cluster2=node_list2
        
#         max_similardis=-float('inf')
#         for i in segs_cluster1:
#             for j in segs_cluster2:
#                 if scores_init.scoremat[i][j]> max_similardis:
#                     max_similardis=scores_init.scoremat[i][j]
#                     seg1=[current_diar[i][3],current_diar[i][4]]
#                     seg2=[current_diar[j][3],current_diar[j][4]]


#     elif selection_criteria == "ideal":
#         segs_cluster1=[]
#         for i in range(len(current_diar)):
#             if current_diar[i][1] in spk_list1:
#                 segs_cluster1.append(i)
#         segs_cluster2=[]
#         for i in range(len(current_diar)):
#             if current_diar[i][1] in spk_list2:
#                 segs_cluster2.append(i)

#         root  ='/lium/buster1/prokopalo/EXPE/icassp_hal_baseline/data/ALLIES/eval'
#         basename = current_diar[0]['show']
#         ref = Diar.read_mdtm(root+'/mdtm/'+basename+'.mdtm')          
        
#         if len(ref)!=len(current_diar):
#             print ("Error ! number of segment in diar are not same ..."+current_diar[0]['show'])
#         dict1 = {}      
#         for indx in segs_cluster1:
#             if len(ref)>indx:
#                 if ref[indx]['cluster'] in dict1.keys():
#                     dict1[ref[indx]['cluster']]+=(ref[indx]['stop']-ref[indx]['start'])
#                 else:
#                     dict1[ref[indx]['cluster']]=(ref[indx]['stop']-ref[indx]['start'])

#         dict2 = {}
#         for indx in segs_cluster2:
#             if len(ref)>indx:
#                 if ref[indx]['cluster'] in dict2.keys():
#                     dict2[ref[indx]['cluster']]+=(ref[indx]['stop']-ref[indx]['start'])
#                 else:
#                     dict2[ref[indx]['cluster']]=(ref[indx]['stop']-ref[indx]['start'])

#         if max(dict2, key = dict2.get) == max(dict1, key = dict1.get):
#             return True , 0
#         else:
#             return False , 0

#         #import pdb
#         #pdb.set_trace()    

#     # For each segments, get the center of the segment in seconds
#     t1 =numpy.mean(seg1)/100
#     t2 =numpy.mean(seg2)/100
    
    
#     # Ask the question to the user
#     message_to_user = MessageToUser(file_info,
#                                     s4d_to_allies(current_diar),
#                                     Request('same', t1, t2))
#     hal, answer = user.validate(message_to_user)
#     return answer.answer,hal


def check_std_change(node_to_check, scores, current_diar, current_vec, link, question):
    """
     Evaluation of the question
    :param node_to_check: 
    :param scores: Scores object with symetric matrix of PLDA scores inside
    :param current_diar: the segmentation that comes out from the automatic pass
    :param current_vec: StatServer of  vectors not YET normalized
    :param link: linkage matrix
    :param question: separation or clustering
    :return: True or False as a proposition of the question to be asked or not
    """
    number_cluster=len(scores.modelset)
    node_list1=get_node_spkeakers(node_to_check[0],number_cluster,link)
    node_list2=get_node_spkeakers(node_to_check[1],number_cluster,link)

    spk_list1=[]
    for n in node_list1:
        spk_list1.append(scores.modelset[n])
    spk_list2=[]
    for n in node_list2:
        spk_list2.append(scores.modelset[n])
        
    segs_cluster_mrg={}
    for i in range(len(current_diar)):
        if current_diar[i][1] in spk_list1 or current_diar[i][1] in spk_list2:
            segs_cluster_mrg[i]=current_vec.stat1[i]
    mrg_situation=numpy.mean(numpy.std(list(segs_cluster_mrg.values()), axis=0))

    segs_cluster1_sep={}
    for i in range(len(current_diar)):
        if current_diar[i][1] in spk_list1:
            segs_cluster1_sep[i]=current_vec.stat1[i]
    std_sep1=numpy.std(list(segs_cluster1_sep.values()), axis=0)
    segs_cluster2_sep={}
    for i in range(len(current_diar)):
        if current_diar[i][1] in spk_list2:
            segs_cluster2_sep[i]=current_vec.stat1[i]
    std_sep2=numpy.std(list(segs_cluster2_sep.values()), axis=0)
    sep_situation=numpy.mean(numpy.mean([std_sep1,std_sep2], axis=0))
    
    if question =="separation":
        question_quality= sep_situation<mrg_situation
    elif question =="clustering":
        question_quality= sep_situation>mrg_situation
    else:
        print("Question type is not valid ...")
        question_quality=True
        
    return question_quality


def check_der(current_diar, current_vec , cluster_list, link, uem, ref):
    
    
    cluster_dict = dict()
    merge = list()
    i = 0

    while i < len(link):

        # the cluster_list of the 2 clusters
        if len(cluster_list)>int(link[i][0]) and len(cluster_list)>int(link[i][1]): 
            c0 = cluster_list[int(link[i][0])]
            c1 = cluster_list[int(link[i][1])]
            information(merge, i, c0, c1, link[i][2])
            if c1 in cluster_dict:
                # c0 is put in c1, and c1 is not empty
                cluster_dict[c1].append(c0)
            else:
                cluster_dict[c1] = [c0]
            if c0 in cluster_dict:
                # remove c0 key
                cluster_dict[c1] += cluster_dict[c0]
                cluster_dict.pop(c0)
            # add the speaker of the new cluster
            cluster_list.append(c1)
            current_diar.rename('cluster', [c0], c1)

        i += 1
    for idx in range(len(current_diar)):
        current_vec.modelset[idx] = current_diar[idx]['cluster']
#     if len(current_diar.unique('cluster'))!=len(set(current_vec.modelset)):
#         import pdb
#         pdb.set_trace()
    
    hyp = s4d_to_allies(current_diar)
    der, fa_rate, miss_rate, conf_rate, time, newspkmap = compute_der(ref, hyp, uem, {}, 0.250)
    
    
    return der, time,current_diar, current_vec


# def active_learning_tree(current_diar,
#                          initial_diar,
#                          current_vec,
#                          scores,
#                          scores_init,
#                          threshold,
#                          user,
#                          file_info,
#                          clustering_method="complete",
#                          selection_method="cluster_center_score",
#                          no_more_clustering = False,
#                          no_more_separation = False,
#                          process_all_nodes = True,
#                          conditional_questioning=False,
#                          uem=None,
#                          ref=None):
#     """

#     :param current_diar: the segmentation that comes out from the automatic pass
#     :param current_vec: StatServer of  vectors not YET normalized
#     :param scores: Scores object with symetric matrix of PLDA scores inside
#     :param threshold: the clustering threshold
#     :param user: user simulation developed in the ALLIES package
#     :param clustering_method: complete
#     :param selection_method: The method of segment selection for asking quastion from human (could be different for seperation and clustering)
#     :param process_all_node: The new stopping creteria for processing whole tree. It changes the limitation of having one line as threshold and lets to consider all branches
#     :param conditional_questioning: Usage of reducing std of cluster as condition for asking question
#     :return:
#     """
#     # Perform HAC on the vectors (that shoiuld produce exactly the same segmentation as the one in current diar
#     ldiar = copy.deepcopy(current_diar)
#     init_diar=copy.deepcopy(initial_diar)
#     lscores = copy.deepcopy(scores)
#     # get the triangular part of the distances
#     distances, th = scores2distance(lscores, threshold)
#     distance_sym = squareform(distances)
    
#     # t = -1.0 * threshold - min
#     # cluster the data
#     link = hac.linkage(distance_sym, method=clustering_method)
    
#     #code for creation beautiful grafs
#     diar_= copy.deepcopy(initial_diar)
#     if False:
#         lscores = copy.deepcopy(scores_init)
#         scores = copy.deepcopy(scores_init)
#         distances, th = scores2distance(lscores, threshold)
#         distance_sym = squareform(distances)
#         link = hac.linkage(distance_sym, method=clustering_method)
#         for i in range (len(diar_)):
#             diar_[i]['cluster']=str(i)


#     tmp = numpy.zeros((link.shape[0], link.shape[1] + 2))
#     tmp[:, :-2] = link
#     tmp[:, -2] = link[:, 2] - th
#     tmp[:, -1] = numpy.abs(link[:, 2] - th)

#     # On trie les regroupements par ordre de proximité au seuil
#     # plus proche en premiere ligne
#     links_to_check = tmp[numpy.argsort(tmp[:, -1])]

#     # On récupère la liste des fusions à faire d'abord.
#     final_links = []
#     for l in link:
#         if l[2] < th:
#             final_links.append(l)

#     # Maintenant on analyse noeud par noeud du plus proche au plus lointain du seuil de fusion
#     # et on gère les deux directions
# #     no_more_clustering = False
# #     no_more_separation = False
    
#     if process_all_nodes:
#         separated_list =[]
#         separation_list = [] 
#         clustering_list = []
#     number_cluster = len(scores.scoremat)
#     complete_list = list(scores.modelset)
    
#     #log the DER
#     link_tmp = copy.deepcopy(final_links)
#     new_diar_tmp = copy.deepcopy(init_diar)
#     cluster_list_tmp = scores.modelset.tolist()
#     der, time = check_der(new_diar_tmp, cluster_list_tmp, link_tmp, uem, ref)
#     print("Initial DER : ",der)
    
#     der_track={"time":time,"der_log":[der],"correction":["initial"],"numQ_clustering":0,"numQ_seperation":0}
    
#     for ltc in links_to_check:
        
#         # Si on cesse dans les deux directions on sort de la boucle de correction
#         if no_more_clustering and no_more_separation:
#             break

#         elif ltc[-2] < 0:  # On est en dessous du seuil, on a déjà regroupé
#             # demande si on sépare: trouve les numéros de deux segments
#             # représentatifs des deux cluster déjà regroupés et pose la question
#             skip = False
#             if process_all_nodes:
#                 if no_more_separation or set(get_node_spkeakers(ltc[0],number_cluster,link)+get_node_spkeakers(ltc[0],number_cluster,link)).issubset(set(separation_list)) :  # On laisse ensemble
#                     skip = True
#             else: 
#                 if no_more_separation:
#                     skip = True
#             if skip or not check_std_change(ltc, scores, init_diar,current_vec,link,"separate",active=conditional_questioning):
#                 pass
#             else:  # On sépare, concretement on retire la fusion de la liste final_link
#                 is_same_speaker,_ = ask_question(ltc, scores,scores_init, init_diar,current_vec, user, file_info,link,selection_criteria=selection_method)
#                 der_track["numQ_seperation"]+=1
#                 if is_same_speaker:
#                     der_track["der_log"].append(der)
#                     der_track["correction"].append("None")
                    
                        
# #                     print("stop separation ...")
#                     if process_all_nodes:
#                         separation_list+= get_node_spkeakers(ltc[0],number_cluster,link)
#                         separation_list+= get_node_spkeakers(ltc[1],number_cluster,link)
#                         if set(complete_list).issubset(set(separation_list)):
#                             no_more_separation = True
#                     else:
#                         no_more_separation = True
    
                       
#                 else:
#                     if process_all_nodes:
#                         separated_list+= get_node_spkeakers(ltc[0],number_cluster,link)
#                         separated_list+= get_node_spkeakers(ltc[1],number_cluster,link)
# #                     print("separate cluster ...")
#                     for ii, fl in enumerate(final_links):
#                         if numpy.array_equal(fl, ltc[:4]):
#                             _ = final_links.pop(ii)
#                     link_tmp = copy.deepcopy(final_links)
#                     new_diar_tmp = copy.deepcopy(init_diar)
#                     cluster_list_tmp = scores.modelset.tolist()
#                     der, time = check_der(new_diar_tmp,cluster_list_tmp,link_tmp, uem, ref)
#                     der_track["der_log"].append(der)
#                     der_track["correction"].append("separate")
#                     print("Separate a cluster. new DER : ",der)
                    
                         

#         elif ltc[-2] > 0:  # On est au dessous du seuil, on n'a pas encore regroupé
#             skip = False
#             if process_all_nodes:
#                 if no_more_clustering or set(get_node_spkeakers(ltc[0],number_cluster,link)+get_node_spkeakers(ltc[0],number_cluster,link)).issubset(set(clustering_list)) :
#                     skip = True
#                 if not set(get_node_spkeakers(ltc[0],number_cluster,link)+get_node_spkeakers(ltc[0],number_cluster,link)).isdisjoint(set(separated_list)):
#                     skip = True 
#             else:
#                 if no_more_clustering:
#                     skip = True
#             if skip or not check_std_change(ltc, scores, init_diar,current_vec,link,"merge",active=conditional_questioning):
#                 pass

#             else:
#                 is_same_speaker,_ = ask_question(ltc, scores,scores_init, init_diar,current_vec, user, file_info,link,selection_criteria=selection_method)  # true
#                 der_track["numQ_clustering"]+=1
#                 if is_same_speaker:  # On fait un nouveau regroupement
# #                     print("group clusters ...")
#                     final_links.append(ltc[:4])
#                     link_tmp = copy.deepcopy(final_links)
#                     new_diar_tmp = copy.deepcopy(init_diar)
#                     cluster_list_tmp = scores.modelset.tolist()
#                     der, time = check_der(new_diar_tmp,cluster_list_tmp,link_tmp, uem, ref)
#                     der_track["der_log"].append(der)
#                     der_track["correction"].append("cluster")
#                     print("Group two clusters. new DER : ",der)
                    
                    
#                 else:
#                     der_track["der_log"].append(der)
#                     der_track["correction"].append("None")
                    
#                     # On arrête de chercher de ce côté
# #                     print("stop grouping ...")
#                     if process_all_nodes:
#                         clustering_list+= get_node_spkeakers(ltc[0],number_cluster,link)
#                         clustering_list+= get_node_spkeakers(ltc[1],number_cluster,link)
#                         if set(complete_list).issubset(set(clustering_list)):
#                             no_more_clustering = True
#                     else:
#                         no_more_clustering= True
    
#     baselink = copy.deepcopy(link)    
    
#     #log the DER
#     link = copy.deepcopy(final_links)
#     new_diar = copy.deepcopy(init_diar)
#     cluster_list = scores.modelset.tolist()
#     der, time = check_der(new_diar,cluster_list,link, uem, ref)
    
#     print("After corrections, final DER :",der)

#     #import pdb
#     #pdb.set_trace()

#     return False, new_diar, current_vec,der_track


def track_correction_process(current_diar,
                             current_vec,
                             scores,
                             temporary_link_list,
                             der_track,
                             correction_type,
                             uem,
                             ref):
    
    der, time, new_diar, new_vec = check_der(current_diar, current_vec, list(scores.modelset), temporary_link_list,uem,ref)
    
    der_track["der_log"].append(der)
    der_track["correction"].append(correction_type)

    logging.critical(f"{correction_type} new DER : {der}")
    return der_track,new_diar, new_vec


def active_learning_tree_two_options(current_diar,
                                     initial_diar,
                                     current_vec,
                                     scores_per_cluster,
                                     scores_per_segment,
                                     threshold,
                                     user,
                                     file_info,
                                     uem=None,
                                     ref=None,
                                     process_all_nodes=True,
                                     clustering_method="complete",
                                     selection_method="ideal",
                                     conditional_questioning=False
                                     ):
    
    # Criteria 1
    if not process_all_nodes:
        return active_learning_tree_1(current_diar,
                                       initial_diar,
                                       current_vec,
                                       scores_per_cluster,
                                       scores_per_segment,
                                       threshold,
                                       user,
                                       file_info,
                                       uem,
                                       ref,
                                       clustering_method=clustering_method,
                                       selection_method=selection_method,
                                       conditional_questioning=conditional_questioning)
    # Criteria 2
    else:
        return active_learning_tree_2(current_diar,
                                       initial_diar,
                                       current_vec,
                                       scores_per_cluster,
                                       scores_per_segment,
                                       threshold,
                                       user,
                                       file_info,
                                       uem,
                                       ref,
                                       clustering_method=clustering_method,
                                       selection_method=selection_method,
                                       conditional_questioning=conditional_questioning)


def active_learning_tree_1(current_diar,
                           initial_diar,
                           current_vec,
                           scores_per_cluster,
                           scores_per_segment,
                           threshold,
                           user,
                           file_info,
                           uem,
                           ref,
                           clustering_method="complete",
                           selection_method="longest",
                           conditional_questioning=False,
                           prioritize_separation2clustering=True):
    """
    Criteria 1: process_all_nodes = False

    :param current_diar: the segmentation that comes out from the automatic pass
    :param initial_diar: the segmentation that comes out from the inital clustering
    :param current_vec: StatServer of  vectors not YET normalized
    :param scores_per_cluster: Scores object of current cluster with symetric matrix of PLDA scores inside
    :param scores_per_segment: intial Scores object of segments with symetric matrix of PLDA scores inside
    :param threshold: the clustering threshold
    :param user: user simulation developed in the ALLIES package
    :param clustering_method: complete
    :param selection_method: The method of segment selection for asking quastion from human (could be different for seperation and clustering)
    :param conditional_questioning: Usage of reducing std of cluster as condition for asking question
    :param prioritize_separation2clustering: in the case of conflict only the seperation will be apply
    :return:
    """
    init_diar=copy.deepcopy(initial_diar)   
    # Get the linkage matrix from the scores
    distances, th = scores2distance(scores_per_cluster, threshold)
    distance_sym = squareform(distances)

    # Perform the clustering
    number_cluster = len(scores_per_cluster.scoremat)
    link = hac.linkage(distance_sym, method=clustering_method)

    # Sort the nodes according to their DELTA to the threshold
    tmp = numpy.zeros((link.shape[0], link.shape[1] + 2))
    tmp[:, :-2] = link
    tmp[:, -2] = link[:, 2] - th
    tmp[:, -1] = numpy.abs(link[:, 2] - th)
    links_to_check = tmp[numpy.argsort(tmp[:, -1])]

    # Initialize the list of link to create
    # We create a new linkage matrix with the clusters which have a distance lower than the threshold
    # This corresponds to the links that must be done if not using any human assistance
    temporary_link_list = []
    for l in link:
        if l[2] < th:
            temporary_link_list.append(l)       # final_links

    # creat der_track dictionary and calculate intial DER 
    der, time,new_diar, new_vec = check_der(current_diar,
                                            current_vec,
                                            list(scores_per_cluster.modelset),
                                            temporary_link_list,
                                            uem, 
                                            ref)
    print("Initial DER : ",der,"(Criteria 1: process_all_nodes = False)")
    der_track={"time":time,"der_log":[der],"correction":["initial"]}
    
    # Check all nodes from the tree
    no_more_clustering = False
    no_more_separation = False
    
    # a list of nodes that have separated to avoid any conflict with clustering 
    # it will be used in case of prioritize_separation2clustering
    separated_list =[] 

    for node in links_to_check:

        # In case we stop exploring the tree
        if no_more_clustering and no_more_separation:
            break

        # Check node below the threshold
        if node[-2] < 0:

            # If conditional_questioning is active, We estimate the quality of the question for this node
            if conditional_questioning:
                # If True we don't ask about this node
                not_suitable_question = check_std_change(node,
                                                         scores_per_cluster,
                                                         init_diar,
                                                         current_vec,
                                                         link,
                                                         "separation")
                # if the node has been labeled as sure enough, we don't ask question to the human
                if not_suitable_question:
                    pass
            
            # If we already decided not tyo explore down the tree
            if no_more_separation:
                pass
            # otherwise as question to the human about this node
            else:
                # Ask the human
                is_same_speaker, _ = ask_question_al(node,
                                                     link,
                                                     scores_per_cluster,
                                                     scores_per_segment,
                                                     init_diar,
                                                     new_diar,
                                                     current_vec,
                                                     user,
                                                     file_info,
                                                     ref,
                                                     selection_criteria = selection_method)


                # if the human validate the node (it has been grouped and it must be)
                # Stop exploring the tree downward
                if is_same_speaker:
                    no_more_separation = True
                    link_tmp = copy.deepcopy(temporary_link_list)
                    diar_tmp = copy.deepcopy(init_diar)
                    der_track,new_diar,new_vec = track_correction_process(diar_tmp,
                                                                  current_vec,
                                                                  scores_per_cluster,
                                                                  link_tmp,
                                                                  der_track,
                                                                  "not_separation",
                                                                  uem,
                                                                  ref)
                # if the human decide to separate the node
                else:
                    #update list to avoid a conflict with clustering
                    separated_list+= get_node_spkeakers(node[0],number_cluster,link)
                    separated_list+= get_node_spkeakers(node[1],number_cluster,link)
                    for ii, fl in enumerate(temporary_link_list):
                        if numpy.array_equal(fl, node[:4]):
                            _ = temporary_link_list.pop(ii)
                    # Record the correction and the DER
                    link_tmp = copy.deepcopy(temporary_link_list)
                    diar_tmp = copy.deepcopy(init_diar)
                    der_track,new_diar,new_vec = track_correction_process(diar_tmp,
                                                                         current_vec,
                                                                         scores_per_cluster,
                                                                         link_tmp,
                                                                         der_track,
                                                                         "separation",
                                                                         uem,
                                                                         ref)

        # Check node above the threshold
        elif node[-2] > 0:

            # If conditional_questioning is active, We estimate the quality of the question for this node
            if conditional_questioning:
                # If True we don't ask about this node
                not_suitable_question = check_std_change(node,
                                                         scores_per_cluster,
                                                         init_diar,
                                                         current_vec,
                                                         link,
                                                         "clustering")
                # if the node has been labeled as sure enough, we don't ask question to the human
                if not_suitable_question:
                    pass
            if prioritize_separation2clustering:
                #In order to avoid any conflict for clustering and separation 
                #and based on the fact that separation gives more gain
                #check if the node is part of a branch that has been clustered before 
                branch1_nodes=get_node_spkeakers(node[0],number_cluster,link)
                branch2_nodes=get_node_spkeakers(node[1],number_cluster,link)
                if not set(branch1_nodes+branch2_nodes).isdisjoint(set(separated_list)):
                    no_more_clustering=True
                    print("In order to avoid conflict and by prioritizing separation, clustering stopped!")
                    pass
            
            # If we already decided not tyo explore up the tree
            if no_more_clustering:
                pass
            # otherwise as question to the human about this node
            else:
                # Ask the human
                is_same_speaker, _ = ask_question_al(node,
                                                     link,
                                                     scores_per_cluster,
                                                     scores_per_segment,
                                                     init_diar,
                                                     new_diar,
                                                     current_vec,
                                                     user,
                                                     file_info,
                                                     ref,
                                                     selection_criteria = selection_method)

                # if the human validate the node (it has not been grouped and it must be)
                if is_same_speaker:
                    temporary_link_list.append(node[:4])
                    # Record the correction and the DER
                    link_tmp = copy.deepcopy(temporary_link_list)
                    diar_tmp = copy.deepcopy(init_diar)
                    der_track,new_diar,new_vec = track_correction_process(diar_tmp,
                                                                         current_vec,
                                                                         scores_per_cluster,
                                                                         link_tmp,
                                                                         der_track,
                                                                         "clustering",
                                                                         uem,
                                                                         ref)

                # Else stop exploring the tree upward
                else:
                    no_more_clustering = True
                    link_tmp = copy.deepcopy(temporary_link_list)
                    diar_tmp = copy.deepcopy(init_diar)
                    der_track,new_diar,new_vec = track_correction_process(diar_tmp,
                                                                         current_vec,
                                                                         scores_per_cluster,
                                                                         link_tmp,
                                                                         der_track,
                                                                         "not_clustering",
                                                                         uem,
                                                                         ref)
    
    return False, new_diar, new_vec, der_track

        
def active_learning_tree_2(current_diar,
                           initial_diar,
                           current_vec,
                           scores_per_cluster,
                           scores_per_segment,
                           threshold,
                           user,
                           file_info,
                           uem,
                           ref,
                           clustering_method="complete",
                           selection_method="longest",
                           conditional_questioning=False,
                           prioritize_separation2clustering=True):
    """
    Criteria 2: process_all_nodes = True

    :param current_diar: the segmentation that comes out from the automatic pass
    :param initial_diar: the segmentation that comes out from the inital clustering
    :param current_vec: StatServer of  vectors not YET normalized
    :param scores_per_cluster: Scores object of current cluster with symetric matrix of PLDA scores inside
    :param scores_per_segment: intial Scores object of segments with symetric matrix of PLDA scores inside
    :param threshold: the clustering threshold
    :param user: user simulation developed in the ALLIES package
    :param clustering_method: complete
    :param selection_method: The method of segment selection for asking quastion from human (could be different for seperation and clustering)
    :param process_all_node: The new stopping creteria for processing whole tree. It changes the limitation of having one line as threshold and lets to consider all branches
    :param conditional_questioning: Usage of reducing std of cluster as condition for asking question
    :param prioritize_separation2clustering: in the case of conflict only the seperation will be apply
    :return:
    """
    init_diar=copy.deepcopy(initial_diar)
    # Get the linkage matrix from the scores
    distances, th = scores2distance(scores_per_cluster, threshold)
    distance_sym = squareform(distances)

    # Perform the clustering
    number_cluster = len(scores_per_cluster.scoremat)
    complete_list = list(scores_per_cluster.modelset)
    link = hac.linkage(distance_sym, method=clustering_method)

    # Sort the nodes according to their DELTA to the threshold
    tmp = numpy.zeros((link.shape[0], link.shape[1] + 2))
    tmp[:, :-2] = link
    tmp[:, -2] = link[:, 2] - th
    tmp[:, -1] = numpy.abs(link[:, 2] - th)
    links_to_check = tmp[numpy.argsort(tmp[:, -1])]

    # Initialize the list of link to create

    # This corresponds to the links that must be done if not using any human assistance
    temporary_link_list = []
    for l in link:
        if l[2] < th:
            temporary_link_list.append(l)       # final_links

    # creat der_track dictionary and calculate intial DER 
    der, time,new_diar, new_vec = check_der(current_diar,
                                            current_vec,
                                            list(scores_per_cluster.modelset),
                                            temporary_link_list,
                                            uem, 
                                            ref)
    print("Initial DER : ",der,"(Criteria 2: process_all_nodes = True)")
    der_track={"time":time,"der_log":[der],"correction":["initial"]}
    
    # Check all nodes from the tree
    no_more_clustering = False
    no_more_separation = False
    
    # a list of nodes that have separated to avoid any conflict with clustering
    # it will be used in case of prioritize_separation2clustering
    separated_list =[]
    stop_separation_list = [] # a list of nodes that have gotten confirmaton for seperation question
    stop_clustering_list = [] # a list of nodes that have gotten confirmaton for clustering question
    
    for node in links_to_check:

        # In case we stop exploring the tree
        if no_more_clustering and no_more_separation:
            break

        # Check node below the threshold
        if node[-2] < 0:

            # If conditional_questioning is active, We estimate the quality of the question for this node
            if conditional_questioning:
                # If True we don't ask about this node
                not_suitable_question = check_std_change(node,
                                                         scores_per_cluster,
                                                         init_diar,
                                                         current_vec,
                                                         link,
                                                         "separation")
                # if the node has been labeled as sure enough, we don't ask question to the human
                if not_suitable_question:
                    pass
            #check if the node is part of a branch that has gotten a confirmation answer before
            branch1_nodes=get_node_spkeakers(node[0],number_cluster,link)
            branch2_nodes=get_node_spkeakers(node[1],number_cluster,link)
            if set(branch1_nodes+branch2_nodes).issubset(set(stop_separation_list)):
                no_more_separation=True
                pass
            
            # If we already decided not tyo explore down the tree
            if no_more_separation:
                pass
            # otherwise as question to the human about this node
            else:
                # Ask the human
                is_same_speaker, _ = ask_question_al(node,
                                                     link,
                                                     scores_per_cluster,
                                                     scores_per_segment,
                                                     init_diar,
                                                     new_diar,
                                                     current_vec,
                                                     user,
                                                     file_info,
                                                     ref,
                                                     selection_criteria = selection_method)

                # if the human validate the node (it has been grouped and it must be)
                if is_same_speaker:
                    stop_separation_list+= get_node_spkeakers(node[0],number_cluster,link)
                    stop_separation_list+= get_node_spkeakers(node[1],number_cluster,link)
                    if set(complete_list).issubset(set(stop_separation_list)):
                        # All down branches have gotten a confirmation answer
                        no_more_separation = True
                    link_tmp = copy.deepcopy(temporary_link_list)
                    diar_tmp = copy.deepcopy(init_diar)
                    der_track,new_diar,new_vec = track_correction_process(diar_tmp,
                                                                         current_vec,
                                                                         scores_per_cluster,
                                                                         link_tmp,
                                                                         der_track,
                                                                         "not_separation",
                                                                         uem,
                                                                         ref)
                # if the human decide to separate the node
                else:
                    #update list to avoid a conflict with clustering
                    separated_list+= get_node_spkeakers(node[0],number_cluster,link)
                    separated_list+= get_node_spkeakers(node[1],number_cluster,link)
                    for ii, fl in enumerate(temporary_link_list):
                        if numpy.array_equal(fl, node[:4]):
                            _ = temporary_link_list.pop(ii)
                    # Record the correction and the DER
                    link_tmp = copy.deepcopy(temporary_link_list)
                    diar_tmp = copy.deepcopy(init_diar)
                    der_track,new_diar,new_vec = track_correction_process(diar_tmp,
                                                                         current_vec,
                                                                         scores_per_cluster,
                                                                         link_tmp,
                                                                         der_track,
                                                                         "separation",
                                                                         uem,
                                                                         ref)

        # Check node above the threshold
        elif node[-2] > 0:

            # If conditional_questioning is active, We estimate the quality of the question for this node
            if conditional_questioning:
                # If True we don't ask about this node
                not_suitable_question = check_std_change(node,
                                                         scores_per_cluster,
                                                         init_diar,
                                                         current_vec,
                                                         link,
                                                         "clustering")
                # if the node has been labeled as sure enough, we don't ask question to the human
                if not_suitable_question:
                    pass
            if prioritize_separation2clustering:
                #In order to avoid any conflict for clustering and separation 
                #and based on the fact that separation gives more gain
                #check if the node is part of a branch that has been clustered before 
                branch1_nodes=get_node_spkeakers(node[0],number_cluster,link)
                branch2_nodes=get_node_spkeakers(node[1],number_cluster,link)
                if not set(branch1_nodes+branch2_nodes).isdisjoint(set(separated_list)):
                    print("In order to avoid conflict and by prioritizing separation, clustering stopped!")
                    pass 
            #check if the node is part of a branch that has gotten a confirmation answer before
            branch1_nodes=get_node_spkeakers(node[0],number_cluster,link)
            branch2_nodes=get_node_spkeakers(node[1],number_cluster,link)
            if set(branch1_nodes+branch2_nodes).issubset(set(stop_clustering_list)):
                no_more_clustering=True
                pass 
            
            # If we already decided not tyo explore up the tree
            if no_more_clustering:
                pass
            # otherwise as question to the human about this node
            else:
                # Ask the human
                is_same_speaker, _ = ask_question_al(node,
                                                     link,
                                                     scores_per_cluster,
                                                     scores_per_segment,
                                                     init_diar,
                                                     new_diar,
                                                     current_vec,
                                                     user,
                                                     file_info,
                                                     ref,
                                                     selection_criteria = selection_method)

                # if the human validate the node (it has not been grouped and it must be)
                if is_same_speaker:
                    temporary_link_list.append(node[:4])
                    # Record the correction and the DER
                    link_tmp = copy.deepcopy(temporary_link_list)
                    diar_tmp = copy.deepcopy(init_diar)
                    der_track,new_diar,new_vec = track_correction_process(diar_tmp,
                                                                         current_vec,
                                                                         scores_per_cluster,
                                                                         link_tmp,
                                                                         der_track,
                                                                         "clustering",
                                                                         uem,
                                                                         ref)

                # Else stop exploring the tree upward
                else:
                    #update list to avoid a conflict with clustering
                    stop_clustering_list+= get_node_spkeakers(node[0],number_cluster,link)
                    stop_clustering_list+= get_node_spkeakers(node[1],number_cluster,link)
                    if set(complete_list).issubset(set(stop_clustering_list)):
                        no_more_clustering = True
                    link_tmp = copy.deepcopy(temporary_link_list)
                    diar_tmp = copy.deepcopy(init_diar)
                    der_track,new_diar,new_vec = track_correction_process(diar_tmp,
                                                                         current_vec,
                                                                         scores_per_cluster,
                                                                         link_tmp,
                                                                         der_track,
                                                                         "not_clustering",
                                                                         uem,
                                                                         ref)

    return False, new_diar, new_vec, der_track


def apply_correction(hyp, scores, c1, c2, method="min"):
    """

    """
    if method == "min":
        hyp = apply_correction_min(hyp, scores, c1, c2)
    elif method == "max":
        hyp = apply_correction_max(hyp, scores, c1, c2)
    elif method == "average":
        hyp = apply_correction_avr(hyp, scores, c1, c2)

    return hyp


def apply_correction_min(hyp,scores,c1,c2):
    c1list=[]
    c2list=[]

    for i in range(len(hyp)):
        if i!=c1 and hyp[i]['cluster'] == hyp[c1]['cluster']:
            c1list.append(i)
        if i!=c2 and hyp[i]['cluster'] == hyp[c2]['cluster']:
            c2list.append(i)

    if len(c1list)==0 and len(c2list)!=0:
        c2_to_c1 = True

    elif len(c2list)==0 and len(c1list)!=0:
        c2_to_c1 = False

    elif len(c1list)==0 and len(c2list)==0:
        c2_to_c1 = True

    else:
        c1dist = []
        c2dist = []
        for i in range(len(c1list)):
            c1dist.append(scores.scoremat[c1][c1list[i]])
        for i in range(len(c2list)):
            c2dist.append(scores.scoremat[c2][c2list[i]])
        if min(c2dist) <  min(c1dist):
            c2_to_c1 = True
        else:
            c2_to_c1 = False

    if c2_to_c1:
        hyp[c1]['cluster'] = hyp[c2]['cluster']
    else:
        hyp[c2]['cluster'] = hyp[c1]['cluster']
    return hyp




def apply_correction_max(hyp,scores,c1,c2):
    c1list=[]
    c2list=[]

    for i in range(len(hyp)):
        if i!=c1 and hyp[i]['cluster'] == hyp[c1]['cluster']:
            c1list.append(i)
        if i!=c2 and hyp[i]['cluster'] == hyp[c2]['cluster']:
            c2list.append(i)

    if len(c1list)==0 and len(c2list)!=0:
        c2_to_c1 = True

    elif len(c2list)==0 and len(c1list)!=0:
        c2_to_c1 = False

    elif len(c1list)==0 and len(c2list)==0:
        c2_to_c1 = True

    else:
        c1dist = []
        c2dist = []
        for i in range(len(c1list)):
            c1dist.append(scores.scoremat[c1][c1list[i]])
        for i in range(len(c2list)):
            c2dist.append(scores.scoremat[c2][c2list[i]])
        if max(c2dist) <  max(c1dist):
            c2_to_c1 = True
        else:
            c2_to_c1 = False

    if c2_to_c1:
        hyp[c1]['cluster'] = hyp[c2]['cluster']
    else:
        hyp[c2]['cluster'] = hyp[c1]['cluster']
    return hyp

def apply_correction_avr(hyp,scores,c1,c2):
    c1list=[]
    c2list=[]

    for i in range(len(hyp)):
        if i!=c1 and hyp[i]['cluster'] == hyp[c1]['cluster']:
            c1list.append(i)
        if i!=c2 and hyp[i]['cluster'] == hyp[c2]['cluster']:
            c2list.append(i)

    if len(c1list)==0 and len(c2list)!=0:
        c2_to_c1 = True

    elif len(c2list)==0 and len(c1list)!=0:
        c2_to_c1 = False

    elif len(c1list)==0 and len(c2list)==0:
        c2_to_c1 = True

    else:
        c1dist = []
        c2dist = []
        for i in range(len(c1list)):
            c1dist.append(scores.scoremat[c1][c1list[i]])
        for i in range(len(c2list)):
            c2dist.append(scores.scoremat[c2][c2list[i]])
        if sum(c2dist)/float(len(c2dist)) <  sum(c1dist)/float(len(c1dist)):
            c2_to_c1 = True
        else:
            c2_to_c1 = False

    if c2_to_c1:
        hyp[c1]['cluster'] = hyp[c2]['cluster']
    else:
        hyp[c2]['cluster'] = hyp[c1]['cluster']
    return hyp



#!/usr/bin/python
import numpy as np
from sidekit import StatServer, Ndx , FactorAnalyser
from sidekit.iv_scoring import fast_PLDA_scoring
import numpy
import copy

def adapt_model(model_iv, g_iv, norm_iv, adaptation_ratio):
    """

    :param model_iv: all vectors (previous + current show)
    :param g_iv: output vectors from the cross_show function
    :param norm_iv: original training vectors (from the training set only)
    :param adaptation_ratio: weight of the out-of domain vectors (between 0 and 1)
    :return:
    """

    vec_dim = g_iv.stat1[0].shape[0]
    adapt_iv, _, __ = keep_recurring_speakers(g_iv, vec_dim, occ_number=2)

    adapt_iv.spectral_norm_stat1(model_iv.sn_mean[:1], model_iv.sn_cov[:1])

    adapted_model_iv = copy.deepcopy(model_iv)
    adapted_model_iv.model_filename = 'model/adapted_model_256_200_100.h5'

    new_plda = FactorAnalyser()
    new_plda.weighted_likelihood_plda(out_stat_server=norm_iv,  # out of domain
                                      in_stat_server=adapt_iv, #
                                      alpha=adaptation_ratio,
                                      rank_f=vec_dim,
                                      nb_iter=10,
                                      scaling_factor=1.,
                                      output_file_name=None,
                                      save_partial=False)

    adapted_model_iv.plda_f = new_plda.F
    adapted_model_iv.plda_sigma = new_plda.Sigma
    adapted_model_iv.plda_mean = new_plda.mean

    return adapted_model_iv



def my_merge(ss_1, ss_2):
    out_sessions = np.unique(ss_1.modelset).shape[0]

    in_sessions = np.unique(ss_2.modelset).shape[0]

    tot_sessions = ss_1.modelset.shape[0] + ss_2.modelset.shape[0]

    flt_iv = StatServer()
    flt_iv.stat0 = np.ones((tot_sessions, 1))
    flt_iv.stat1 = np.ones((tot_sessions, ss_1.stat1.shape[1]))
    flt_iv.modelset = np.empty(tot_sessions, dtype="|O")
    flt_iv.segset = np.empty(tot_sessions, dtype="|O")
    flt_iv.start = np.empty(tot_sessions, dtype="|O")
    flt_iv.stop = np.empty(tot_sessions, dtype="|O")

    for idx in range(ss_1.modelset.shape[0]):
        flt_iv.stat1[idx, :] = ss_1.stat1[idx, :]
        flt_iv.modelset[idx] = str(ss_1.modelset[idx])
        flt_iv.segset[idx] = ss_1.segset[idx]
        flt_iv.start[idx] = ss_1.start[idx]
        flt_iv.stop[idx] = ss_1.stop[idx]

    for idx2 in range(ss_2.modelset.shape[0]):
        idx = idx2 + ss_1.modelset.shape[0]
        flt_iv.stat1[idx, :] = ss_2.stat1[idx2, :]
        flt_iv.modelset[idx] = str(ss_2.modelset[idx2])
        flt_iv.segset[idx] = ss_2.segset[idx2]
        flt_iv.start[idx] = ss_2.start[idx2]
        flt_iv.stop[idx] = ss_2.stop[idx2]

    return flt_iv


def keep_recurring_speakers(iv_ss, rank_F, occ_number=3, filter_by_name=False):
    unique, counts = numpy.unique(iv_ss.modelset, return_counts=True)
    tot_sessions = 0
    kept_idx = []
    for model in range(unique.shape[0]):
        if counts[model] >= occ_number \
                and ((not (unique[model]).startswith("speaker")
                     and not (unique[model]).startswith("presentat")
                     and not (unique[model]).startswith("sup+")
                     and not (unique[model]).startswith("voix")
                     and not (unique[model]).startswith("publicitaire")
                     and not (unique[model]).startswith("locuteur")
                     and not (unique[model]).startswith("inconnu")
                     and not (unique[model]).startswith("traduct")
                     and not ("#" in (unique[model]))
                     and "_" in (unique[model])
                     and filter_by_name)
                     or (not filter_by_name)):
            kept_idx = numpy.append(kept_idx, numpy.where(iv_ss.modelset == unique[model]))
            tot_sessions += counts[model]
    flt_iv = StatServer()
    flt_iv.stat0 = numpy.ones((tot_sessions, 1))
    flt_iv.stat1 = numpy.ones((tot_sessions, rank_F))
    flt_iv.modelset = numpy.empty(tot_sessions, dtype="|O")
    flt_iv.segset = numpy.empty(tot_sessions, dtype="|O")
    flt_iv.start = numpy.empty(tot_sessions, dtype="|O")
    flt_iv.stop = numpy.empty(tot_sessions, dtype="|O")

    i = 0
    for model in range(unique.shape[0]):
        if counts[model] >= occ_number \
                and ((not (unique[model]).startswith("speaker")
                     and not (unique[model]).startswith("presentat")
                     and not (unique[model]).startswith("sup+")
                     and not (unique[model]).startswith("voix")
                     and not (unique[model]).startswith("publicitaire")
                     and not (unique[model]).startswith("locuteur")
                     and not (unique[model]).startswith("inconnu")
                     and not (unique[model]).startswith("traduct")
                     and not ("#" in (unique[model]))
                     and "_" in (unique[model])
                     and filter_by_name)
                     or (not filter_by_name)):
            for ivector in numpy.where(iv_ss.modelset == unique[model])[0]:
                flt_iv.stat1[i, :] = iv_ss.stat1[ivector, :]
                flt_iv.modelset[i] = iv_ss.modelset[ivector]
                flt_iv.segset[i] = iv_ss.segset[ivector]
                flt_iv.start[i] = iv_ss.start[ivector]
                flt_iv.stop[i] = iv_ss.stop[ivector]
                i += 1
    unique = numpy.unique(flt_iv.modelset)
    return flt_iv, unique.shape[0], kept_idx




# -* coding: utf-8 -*-

#################################################################################
# This file is part of EVALLIES.                                                #
#                                                                               #
# EVALLIES is a python package for lifelong learning speaker diarization.       #
# Home page: https://git-lium.univ-lemans.fr/Larcher/evallies                   #
#                                                                               #
# EVALLIES is free software: you can redistribute it and/or modify              #
# it under the terms of the GNU LLesser General Public License as               #
# published by the Free Software Foundation, either version 3 of the License,   #
# or (at your option) any later version.                                        #
#                                                                               #
# EVALLIES is distributed in the hope that it will be useful,                   #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 #
# GNU Lesser General Public License for more details.                           #
#                                                                               #
# You should have received a copy of the GNU Lesser General Public License      #
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.              #
#                                                                               #
#################################################################################
"""
Copyright 2020-2021 Anthony Larcher, Meysam Shamsi & Yevhenii Propkopalo

    :mod:`lium_baseline.system`

"""


import copy
import decimal
import fastcluster
import h5py
import logging
import numpy
import os
import pandas
import pickle
import sidekit
import soundfile
import s4d
import sys
import torch
import torch
import yaml

from scipy.cluster import hierarchy as hac

from scipy.spatial.distance import squareform

from .utils import add_show_in_id
from .utils import remove_show_from_id
from .utils import s4d_to_allies

from .utils import keep_recurring_speakers

from ..user_simulation import MessageToUser
from ..user_simulation import Request

from .interactive import apply_correction
from .interactive import active_learning_tree_two_options

#from .plda_adapt import adapt_model

def concat_statservers(ss_1, ss_2):
    out_sessions = numpy.unique(ss_1.modelset).shape[0]

    in_sessions = numpy.unique(ss_2.modelset).shape[0]

    tot_sessions = ss_1.modelset.shape[0] + ss_2.modelset.shape[0]

    flt_iv = sidekit.StatServer()
    flt_iv.stat0 = numpy.ones((tot_sessions, 1))
    flt_iv.stat1 = numpy.ones((tot_sessions, ss_1.stat1.shape[1]))
    flt_iv.modelset = numpy.empty(tot_sessions, dtype="|O")
    flt_iv.segset = numpy.empty(tot_sessions, dtype="|O")
    flt_iv.start = numpy.empty(tot_sessions, dtype="|O")
    flt_iv.stop = numpy.empty(tot_sessions, dtype="|O")

    for idx in range(ss_1.modelset.shape[0]):
        flt_iv.stat1[idx, :] = ss_1.stat1[idx, :]
        flt_iv.modelset[idx] = str(ss_1.modelset[idx])
        flt_iv.segset[idx] = ss_1.segset[idx]
        flt_iv.start[idx] = ss_1.start[idx]
        flt_iv.stop[idx] = ss_1.stop[idx]

    for idx2 in range(ss_2.modelset.shape[0]):
        idx = idx2 + ss_1.modelset.shape[0]
        flt_iv.stat1[idx, :] = ss_2.stat1[idx2, :]
        flt_iv.modelset[idx] = str(ss_2.modelset[idx2])
        flt_iv.segset[idx] = ss_2.segset[idx2]
        flt_iv.start[idx] = ss_2.start[idx2]
        flt_iv.stop[idx] = ss_2.stop[idx2]

    return flt_iv


def rename_models(within_iv, within_diar, existing_models):
    """
    """
    within_models = dict()
    idx = 0

    for mod in set(within_iv.modelset):
        while f"Spk{idx}" in existing_models:
            idx += 1
        within_models[mod] = f"Spk{idx}"
        idx += 1

    within_iv.modelset = within_iv.modelset.astype('U50')
    for idx, mod in enumerate(within_iv.modelset):
        within_iv.modelset[idx] = within_models[mod]


    for idx, seg in enumerate(within_diar.segments):
        within_diar.segments[idx]['cluster'] = within_models[seg['cluster']]

    return within_iv, within_diar


def perform_iv_seg(sn_mean,
                   sn_cov,
                   plda_mean,
                   plda_f,
                   plda_sigma,
                   th_W,
                   initial_diar,
                   within_iv,
                   hac_method,
                   precomputed_scores=None):
    """

    """

    # Compute scores for each pair of segment to be used in human assisted correction
    # It provides information


    within_iv_mean = within_iv.mean_stat_per_model()

    # Normalize i-vectors
    within_iv_mean.spectral_norm_stat1(sn_mean[:1], sn_cov[:1])
    if within_iv_mean.modelset.shape[0] > 1:
        if precomputed_scores is None:
            # Compute scores
            ndx = sidekit.Ndx(models=within_iv_mean.modelset, testsegs=within_iv_mean.modelset)
            scores = sidekit.iv_scoring.fast_PLDA_scoring(within_iv_mean, within_iv_mean, ndx,
                                                          plda_mean,
                                                          plda_f,
                                                          plda_sigma,
                                                          p_known=0.0,
                                                          scaling_factor=1.0,
                                                          check_missing=False)
            scores.scoremat = 0.5 * (scores.scoremat + scores.scoremat.transpose())
        else:
            scores = precomputed_scores
            scores.scoremat = 0.5 * (scores.scoremat + scores.scoremat.transpose())

        # Run HAC clustering
        new_diar, cluster_dict, merge = s4d.clustering.hac_iv.hac_iv(initial_diar, scores, threshold=th_W, method=hac_method)
        

        # Update the model names of i-vector
        # reverse the cluster dict:
        update_dict = dict()
        for k, v in cluster_dict.items():
            for _v in v:
                update_dict[_v] = k
        
        for idx, mod in enumerate(within_iv.modelset):
            if mod in update_dict:
                within_iv.modelset[idx] = update_dict[mod]

    else:
        new_diar = initial_diar
        scores = None 

    return new_diar, within_iv, scores


def cross_show(model, model_cfg, previous_iv, previous_diar, within_iv, within_diar, th_x, incremental=True):
    """
    """
    # Make sure model names are not overlapping (change within show models if needed)
    within_iv, within_diar = rename_models(within_iv, within_diar, existing_models=previous_iv.modelset)
    # backup previous_iv and within_iv
    within_iv_backup = copy.deepcopy(within_iv)
    previous_iv_backup = copy.deepcopy(previous_iv)

    # get the mean_per_model for previous and within
    within_iv_mean = within_iv.mean_stat_per_model()
    previous_iv_mean = previous_iv.mean_stat_per_model()

    # merge the mean_per_model for previous and within
    ll_iv_mean = concat_statservers(previous_iv_mean, within_iv_mean)

    # compute scores, constrain and run HAC clustering
    # Normalize all
    ll_iv_mean.spectral_norm_stat1(model["model_iv"].sn_mean[:1], model["model_iv"].sn_cov[:1])

    # Compute the score matrix
    ndx = sidekit.Ndx(models=ll_iv_mean.modelset, testsegs=ll_iv_mean.modelset)
    scores = sidekit.iv_scoring.fast_PLDA_scoring(ll_iv_mean, ll_iv_mean, ndx,
                                                  model["model_iv"].plda_mean,
                                                  model["model_iv"].plda_f,
                                                  model["model_iv"].plda_sigma,
                                                  p_known=0.0,
                                                  scaling_factor=1.0,
                                                  check_missing=False)
    scores.scoremat = -0.5 * (scores.scoremat + scores.scoremat.transpose())

    # Constrain the scores to forbid any new clustering between previous shows
    lowest_distance = numpy.min(scores.scoremat)
    for iv_idx in range(previous_iv_mean.modelset.shape[0]):
        for iv_jdx in range(previous_iv_mean.modelset.shape[0]):
            if previous_iv_mean.modelset[iv_idx] == previous_iv_mean.modelset[iv_jdx]:
                scores.scoremat[iv_idx, iv_jdx] = lowest_distance - 1
            else:
                scores.scoremat[iv_idx, iv_jdx] = numpy.inf


    # Add to keep the within show clustering
    for iv_idx in range(previous_iv_mean.modelset.shape[0], ll_iv_mean.modelset.shape[0]):
        for iv_jdx in range(previous_iv_mean.modelset.shape[0], ll_iv_mean.modelset.shape[0]):
            if ll_iv_mean.modelset[iv_idx] == ll_iv_mean.modelset[iv_jdx]:
                scores.scoremat[iv_idx, iv_jdx] = lowest_distance - 1
            else:
                scores.scoremat[iv_idx, iv_jdx] = numpy.inf
    # fin modif

    # Modify the scores between old shows and the current one accordingly
    numpy.fill_diagonal(scores.scoremat, numpy.inf)
    for idx in range(previous_iv_mean.modelset.shape[0], ll_iv_mean.modelset.shape[0]):
        for jdx in range(previous_iv_mean.modelset.shape[0]):
            scores.scoremat[idx, jdx] = numpy.max(scores.scoremat[idx, numpy.where(previous_iv_mean.modelset == previous_iv_mean.modelset[jdx])])
            scores.scoremat[jdx, idx] = scores.scoremat[idx, jdx]
    numpy.fill_diagonal(scores.scoremat, 0)

    numpy.fill_diagonal(scores.scoremat, numpy.inf)

    #with open("matrix_ cross_show.txt", 'w') as fh:
    #    for ii in range(scores.scoremat.shape[0]):
    #        for jj in range(scores.scoremat.shape[1]):
    #            fh.write(f"{scores.scoremat[ii, jj]:3.1f},\t")
    #        fh.write("\n")

    mini0 = numpy.min(scores.scoremat) - 1
    scores.scoremat -= mini0

    modelset_seg_idx = dict()
    for seg in previous_diar.segments:
        modelset_seg_idx[seg['cluster']] = numpy.where(ll_iv_mean.modelset == seg['cluster'])[0]

    numpy.fill_diagonal(scores.scoremat, 0.0)
    squareform_plda = squareform(scores.scoremat)
    #Z = fastcluster.linkage(squareform_plda, method='complete', metric='euclidean', preserve_input=True)
    Z = fastcluster.linkage(squareform_plda, method='complete', preserve_input=True)

    """
    if we want to set the final number of cluster (approach DiHard)
    the next line should be replaced with

    final_cluster_nb = 12
    T = hac.fcluster(Z, t=final_cluster_nb, criterion="maxclust")

    Beware, final_cluster_nb should be higher than than the number of cluster existing in previous_iv_mean
    """
    T = hac.fcluster(Z, th_x - mini0, 'distance')

    # DSon't allow to modify the names of previously existing clusters
    # Create a dictionnary with old_model_name as key and new_luster as value
    cluster_dict = dict()
    clusters_by_index = dict()
    for ii in range(T.shape[0]):
        if T[ii] not in clusters_by_index:
            clusters_by_index[T[ii]] = ll_iv_mean.modelset[ii]
        cluster_dict[ll_iv_mean.modelset[ii]] = clusters_by_index[T[ii]]

    # concatenate previous_iv et within_iv
    new_previous_iv = concat_statservers(previous_iv_backup, within_iv_backup)
    new_previous_diar = copy.deepcopy(previous_diar)
    new_previous_diar.segments += within_diar.segments

    # Modify the model names for i-vectors
    for ii, mod in enumerate(new_previous_iv.modelset):
        new_previous_iv.modelset[ii] = cluster_dict[mod]

    # modify the model names for diar
    for ii, seg in enumerate(new_previous_diar.segments):
        new_previous_diar.segments[ii]['cluster'] = cluster_dict[seg['cluster']]

    for ii, seg in enumerate(within_diar.segments):
        within_diar.segments[ii]['cluster'] = cluster_dict[seg['cluster']]

    return new_previous_iv, new_previous_diar, within_diar

#def allies_load_seg(model_cfg, th, show):
#
#    current_diar = s4d.Diar.read_seg(f"{model_cfg['tmp_dir']}/seg/within_{th}/{show}.seg")
#    current_vec = sidekit.StatServer(f"{model_cfg['tmp_dir']}/seg/within_{th}/{show}_vec.h5")
#
#    return current_diar, current_vec


def allies_initial_training(system_config, train_data):
    """
    Train the initial Diarization sys 311     with open(system_config, 'r') as fh:
 312         model_cfg = yaml.load(fh, Loader=yaml.FullLoader)tem
    """
    # Load the config file
    with open(system_config, 'r') as fh:
        model_cfg = yaml.load(fh, Loader=yaml.FullLoader)

    if model_cfg["model"]["type"] == "lium_iv":
        model, model_cfg = lium_iv_initial_training(model_cfg, train_data)

    elif model_cfg["model"]["type"] == "lium_xv":
        model, model_cfg = lium_xv_initial_training(model_cfg, train_data)

    elif model_cfg["model"]["type"] == "lium_sliding_iv":
        model = None

    elif model_cfg["model"]["type"] == "lium_sliding_xv":
        model = None

    return model, model_cfg



def lium_sliding_xv(model_cfg, train_data):
    pass



def lium_iv_initial_training(model_cfg, train_data):
    """
    Extract features and store them in tmp folder
    Train UBM
    Train TV
    Extract i-vectors
    Train PLDA

    :param model_cfg: a dictionnary describing the system parameters 
    :param train_data: a Data generator used to access the data
    """

    fe = sidekit.FeaturesExtractor(**model_cfg["model"]["feature_extractor"])
    fs_acoustic = sidekit.FeaturesServer(feature_filename_structure=model_cfg['tmp_dir'] + "/feat/{}.h5",
                                         **model_cfg["model"]["vectors"]["ivectors"]["feature_server"])

    train_diar = s4d.diar.Diar()
    train_diar.add_attribut(new_attribut='gender', default='U')
    name_dict = {}
    # iterate on train_data
    if not os.path.isdir(f"{model_cfg['tmp_dir']}/feat"):
        os.makedirs(f"{model_cfg['tmp_dir']}/feat")

    for idx, (show, file_info, uem, ref, filename) in enumerate(train_data):

        if not os.path.isfile(f"{model_cfg['tmp_dir']}/feat/{show}.h5"):
            fe.save(show,
                    channel=0,
                    input_audio_filename=filename,
                    output_feature_filename=f"{model_cfg['tmp_dir']}/feat/{show}.h5")

        # Use the ref info to fill the Diar object
        for spk, start, end in zip(ref.speaker, ref.start_time, ref.end_time):
            train_diar.append(show=show,
                              cluster= spk,
                              start=int(round(float(start) * 100.)),
                              stop=int(round(float(end) * 100.))
                              )

    print(f'Training data contains:')
    print("\t" + f'{len(train_diar.unique("show"))} files')
    print("\t" + f'{len(train_diar.unique("cluster"))} speakers')
    print("\t" + f'{len(train_diar)} segments')
    name_dict[show] = int(idx)


    ######################################
    # Train UBM
    ######################################
    ubm_diar = train_diar.copy_structure()
    ubm_diar.segments = train_diar.segments[::20]
    ubm_idmap = ubm_diar.id_map()

    if not os.path.isfile(f"{model_cfg['tmp_dir']}/model/ubm.h5"):
        ubm = sidekit.Mixture()
        logging.critical("Start UBM training with {} segments".format(len(ubm_idmap.leftids)))
        ubm.EM_split(fs_acoustic,
                     ubm_idmap,
                     model_cfg['model']['vectors']['ivectors']['distrib_nb'],
                     num_thread=model_cfg['nb_thread'],
                     save_partial=False)
        ubm.write(f"{model_cfg['tmp_dir']}/model/ubm.h5")
    else:
        logging.critical("Load UBM from file")
        ubm = sidekit.Mixture(f"{model_cfg['tmp_dir']}/model/ubm.h5")
    logging.critical("\t Done")

    # Get the size of features output by the FeatureServer
    feature_size = ubm.dim()

    ######################################
    # Train TV
    ######################################
    long_seg_diar = copy.deepcopy(train_diar)
    short_seg_diar = long_seg_diar.filter("duration", "<", 100)
    long_seg_diar = long_seg_diar.filter("duration", ">=", 100)
    tv_idmap = long_seg_diar.id_map()

    if not os.path.isfile(f"{model_cfg['tmp_dir']}/model/TV.h5"):

        # Accumulate sufficient statistics for the training data
        logging.critical("Create StatServer with {} segments".format(len(tv_idmap.leftids)) )
        tv_stat = sidekit.StatServer(tv_idmap,
                                     distrib_nb=ubm.get_distrib_nb(),
                                     feature_size=feature_size)

        tv_stat.accumulate_stat(ubm=ubm,
                                feature_server=fs_acoustic,
                                seg_indices=range(tv_stat.segset.shape[0]),
                                num_thread=model_cfg['nb_thread']
                                )

        tv_stat = add_show_in_id(tv_stat)
        tv_stat.write(f"{model_cfg['tmp_dir']}/model/tv_stat.h5")
        logging.critical("\t Done")

        # Sufficient statistics are passed to the FactorAnalyser via an HDF5 file handler that stays in memory
        tv_stat_h5f = h5py.File(f'{model_cfg["tmp_dir"]}/tmp_tv_stat.h5', 'a', backing_store=False, driver='core')
        tv_stat.to_hdf5(tv_stat_h5f)
        #tv_stat.write(f'{model_cfg["tmp_dir"]}/tv_stat.h5', prefix='', mode='w')
        tv_fa = sidekit.FactorAnalyser()
        logging.critical("Start TV training with {} segments".format(len(tv_stat.modelset)))
        tv_fa.total_variability(tv_stat_h5f,
                                ubm,
                                tv_rank=model_cfg['model']['vectors']['size'],
                                nb_iter=10,
                                min_div=True,
                                tv_init=None,
                                batch_size=300,
                                save_init=False,
                                output_file_name=None,
                                num_thread=model_cfg['nb_thread'])
        tv_fa.write(f"{model_cfg['tmp_dir']}/model/TV.h5")
    else:
        logging.critical("Load TV from file")
        tv_fa = sidekit.FactorAnalyser(f"{model_cfg['tmp_dir']}/model/TV.h5")
    logging.critical("\t Done")

    ######################################
    # Extract I-vectors and train PLDA
    ######################################
    if not os.path.isfile(f"{model_cfg['tmp_dir']}/model/training_ivectors.h5"):

        if not os.path.isfile(f"{model_cfg['tmp_dir']}/model/short_seg_stat.h5"):
            # Accumulate sufficient statistics for the short segment from training data
            short_seg_idmap = short_seg_diar.id_map()
            short_seg_stat = sidekit.StatServer(short_seg_idmap,
                                                distrib_nb=ubm.get_distrib_nb(),
                                                feature_size= feature_size)
            logging.critical("Start computing short segment statistics: {}".format(len(short_seg_stat.modelset)))

            short_seg_stat.accumulate_stat(ubm=ubm,
                                           feature_server=fs_acoustic,
                                           seg_indices=range(short_seg_stat.segset.shape[0]),
                                           num_thread=model_cfg['nb_thread']
                                           )
            logging.critical("short iv stat ok")
            # Rename in order to get one i-vector per segment
            #tv_stat = add_show_in_id(tv_stat)
            short_seg_stat = add_show_in_id(short_seg_stat)

            short_seg_stat.write(f"{model_cfg['tmp_dir']}/model/short_seg_stat.h5")



        logging.critical("Start extracting i-vectors")
        long_iv = tv_fa.extract_ivectors(ubm, f"{model_cfg['tmp_dir']}/model/tv_stat.h5", num_thread=model_cfg['nb_thread'])
        short_iv = tv_fa.extract_ivectors(ubm, f"{model_cfg['tmp_dir']}/model/short_seg_stat.h5",  num_thread=model_cfg['nb_thread'])

        current_iv = long_iv.merge(long_iv, short_iv)

        current_iv.write(f"{model_cfg['tmp_dir']}/model/training_ivectors.h5")
    else:
        logging.critical("Load IV from file")
        current_iv = sidekit.StatServer(f"{model_cfg['tmp_dir']}/model/training_ivectors.h5")
    logging.critical("\tI-vectors OK")


    # After taking the mean per speaker per file, we rename the clusters in order
    # to get one mean i-vector per speaker per file but same cluster_ID across shows
    current_iv_mean = current_iv.mean_stat_per_model()
    current_iv_mean = remove_show_from_id(current_iv_mean)

    # Normalize i-vectors
    norm_iv = copy.deepcopy(current_iv_mean)
    norm_mean, norm_cov = norm_iv.estimate_spectral_norm_stat1(1, 'sphNorm')
    norm_iv, n_recc, _ = keep_recurring_speakers(norm_iv, rank_F=model_cfg['model']['vectors']['size'], occ_number=2, filter_by_name=True)

    norm_iv.spectral_norm_stat1(norm_mean, norm_cov)
    logging.critical("start PLDA training")
    #plda_fa = sidekit.FactorAnalyser() Train  PLDA
    plda_fa = sidekit.FactorAnalyser()
    plda_fa.plda(norm_iv,
                 rank_f=model_cfg['model']['classifier']['plda_rank'],
                 nb_iter=20,
                 scaling_factor=1.,
                 output_file_name=f"{model_cfg['tmp_dir']}/model/plda",
                 save_partial=False)
    logging.critical("PLDA ok")

    model_iv = s4d.ModelIV()
    model_iv.ubm = ubm
    model_iv.tv = tv_fa.F
    model_iv.tv_mean = tv_fa.mean
    model_iv.tv_sigma = tv_fa.Sigma
    model_iv.sn_mean = norm_mean
    model_iv.sn_cov = norm_cov
    model_iv.plda_mean = plda_fa.mean
    model_iv.plda_f = plda_fa.F
    model_iv.plda_g = None
    model_iv.plda_sigma = plda_fa.Sigma
    model_iv.ivectors = current_iv
    model_iv.scores = None
    model_iv.nb_thread = 1

    model = dict()
    model["model_iv"] = model_iv
    model["global_diar"] = train_diar
    pickle.dump(model, open(f"{model_cfg['tmp_dir']}/model/model_allies_baseline_{model_cfg['model']['vectors']['type']}v.p", "wb" ))

    return model, model_cfg


def lium_vad_training(model_cfg, train_data):
    """
    Select data for VAD training
    Train the model
    return the model

    :param model_cfg:
    :param train_data:
    :return:
    """
    # Prepare the training data
     #IL SUFFIT DE GENERER LA LISTE DE FICHIERS ET DE LA STOCKER DANS TMP
    training_file_list = []
    for idx, (show, file_info, uem, ref, filename) in enumerate(train_data):
        training_file_list.append(show)
    with open("list/vad_list.lst", "w") as fh:
        fh.write("\n".join(training_file_list))

    # Iterative training
    s4d.nnet.seqTrain(dataset_yaml=model_cfg['model']['vad']['db_yaml'],
                      model_yaml=model_cfg['model']['vad']['model_yaml'],
                      epochs=model_cfg['model']['vad']['epochs'],
                      lr=model_cfg['model']['vad']['lr'],
                      patience=2,
                      tmp_model_name=model_cfg['tmp_dir']+"/model/tmp_vad",
                      best_model_name=model_cfg['tmp_dir']+"/model/best_vad",
                      multi_gpu=True,
                      opt=model_cfg['model']['vad']['opt'],
                      log_interval=10,
                      num_thread=model_cfg['nb_thread']
                      )


def lium_xv_initial_training(model_cfg, train_data):
    """

    """

    if not os.path.isdir(f"{model_cfg['tmp_dir']}/model"):
        os.makedirs(f"{model_cfg['tmp_dir']}/model")

    # Train the VAD and save to disk
    train_data_vad = copy.deepcopy(train_data)
    if not os.path.isfile(model_cfg['tmp_dir'] + "/model/best_vad.pt"):
        lium_vad_training(model_cfg, train_data_vad)

    train_diar = s4d.diar.Diar()
    df_data = dict()
    df_data['speaker_idx'] = []
    df_data['database'] = []
    df_data['speaker_id'] = []
    df_data['start'] = []
    df_data['duration'] = []
    df_data['file_id'] = []
    df_data['gender'] = []
    train_diar.add_attribut(new_attribut='gender', default='U')
    name_dict = {}

    # iterate on train_data
    spk_idx = 0
    speaker_idx = dict()
    for idx, (show, file_info, uem, ref, filename) in enumerate(train_data):

        # Use the ref info to fill the Diar object
        # fill a pandas.DataFrame at the sale tim to select data for X-vector extractor training
        for spk, start, end in zip(ref.speaker, ref.start_time, ref.end_time):
            if spk not in speaker_idx:
                speaker_idx[spk] = spk_idx
                spk_idx += 1

            train_diar.append(show=show,
                              cluster= spk,
                              start=int(round(float(start) * 100.)),
                              stop=int(round(float(end) * 100.)))

            df_data["speaker_idx"].append(speaker_idx[spk])
            df_data["speaker_id"].append(spk)
            df_data["start"].append(start)
            df_data["duration"].append(end - start)
            df_data["file_id"].append(show)
            df_data["database"].append("allies_train")
            df_data["gender"].append('u')

    print(f'Training data contains:')
    print("\t" + f'{len(train_diar.unique("show"))} files')
    print("\t" + f'{len(train_diar.unique("cluster"))} speakers')
    print("\t" + f'{len(train_diar)} segments')

    df = pandas.DataFrame(df_data)
    df = df.loc[df['duration'] > model_cfg['model']['vectors']['xvectors']['duration']]
    vc = df.speaker_idx.value_counts()
    selected_df = df[df.speaker_idx.isin(vc.index[vc.values >= model_cfg['model']['vectors']['xvectors']['min_nb_sessions']])]
    speaker_new_idx = dict(zip(list(selected_df.speaker_idx.unique()), numpy.arange(len(list(selected_df.speaker_idx.unique())))))
    selected_df = selected_df.reset_index(drop=True)
    for idx in range(len(selected_df)):
        selected_df.at[idx, 'speaker_idx'] = speaker_new_idx[selected_df.at[idx, 'speaker_idx']]
    
    selected_df.to_csv("list/xvector.csv", index=False)

    training_speaker_nb = len(selected_df.speaker_idx.unique())

    model_cfg["model"]["speaker_nb"] = training_speaker_nb
    #model_cfg["model"]["speaker_nb"] = 111
    if not os.path.isfile(model_cfg['tmp_dir']+"/model/best_xtractor.pt"):
        sidekit.nnet.xtrain(speaker_number=training_speaker_nb,
                            dataset_yaml=model_cfg['model']['vectors']['xvectors']['db_yaml'],
                            epochs=model_cfg['model']['vectors']['xvectors']['epochs'],
                            lr=model_cfg['model']['vectors']['xvectors']['lr'],
                            model_yaml=model_cfg['model']['vectors']['xvectors']['xtractor_yaml'],
                            model_name=None,
                            loss=model_cfg['model']['vectors']['xvectors']["loss"],
                            tmp_model_name=model_cfg['tmp_dir']+"/model/tmp_xtractor",
                            patience=50,
                            opt=model_cfg['model']['vectors']['xvectors']['opt'],
                            best_model_name=model_cfg['tmp_dir']+"/model/best_xtractor",
                            multi_gpu=True,
                            num_thread=model_cfg['nb_thread'])

    # Extract x-vectors
    train_diar = train_diar.filter("duration", ">=", 100)

    #############
    # Select speakers for PLDA training
    # Keep only sessions longer than model_cfg['model']['vectors']['xvectors']['duration']
    # And speakers having at least model_cfg['model']['vectors']['xvectors']['min_nb_sessions'] sessions
    #############
    train_diar = train_diar.filter("duration", ">=", int(model_cfg['model']['vectors']['xvectors']['duration'] * 100))

    training_idmap = train_diar.id_map()
    training_idmap.start = training_idmap.start * 160
    training_idmap.stop = training_idmap.stop * 160

    if not os.path.isfile(f"{model_cfg['tmp_dir']}/model/training_xv.h5"):
        training_xv = sidekit.nnet.xvector.extract_embeddings(idmap_name=training_idmap,
                                                              model_filename=model_cfg['tmp_dir']+"/model/best_xtractor.pt",
                                                              data_root_name=train_data.root_folder + '/wav/',
                                                              device=torch.device("cuda:0"),
                                                              transform_pipeline=model_cfg['model']['vectors']['xvectors']["transforms"])
        training_xv.write(f"{model_cfg['tmp_dir']}/model/training_xv.h5") 

    training_xv = sidekit.StatServer(f"{model_cfg['tmp_dir']}/model/training_xv.h5")
    # Compute normalization parameters
    training_xv, n_recc, _ = keep_recurring_speakers(training_xv,
                                                     rank_F=model_cfg['model']['vectors']['size'],
                                                     occ_number=model_cfg['model']['vectors']['xvectors']['min_nb_sessions'],
                                                     filter_by_name=True)
    training_xv = add_show_in_id(training_xv)
    training_xv_mean = training_xv.mean_stat_per_model()
    training_xv_mean = remove_show_from_id(training_xv_mean)

    norm_xv = copy.deepcopy(training_xv_mean)
    norm_mean, norm_cov = norm_xv.estimate_spectral_norm_stat1(1, 'sphNorm')
    # Train PLDA
    norm_xv.spectral_norm_stat1(norm_mean, norm_cov)
    plda_fa = sidekit.FactorAnalyser()
    plda_fa.plda(norm_xv,
                 rank_f=model_cfg['model']['classifier']['plda_rank'],
                 nb_iter=50,
                 scaling_factor=1.,
                 output_file_name=f"{model_cfg['tmp_dir']}/model/plda",
                 save_partial=False)


    model_iv = s4d.ModelIV()
    model_iv.sn_mean = norm_mean
    model_iv.sn_cov = norm_cov
    model_iv.plda_mean = plda_fa.mean
    model_iv.plda_f = plda_fa.F
    model_iv.plda_g = None
    model_iv.plda_sigma = plda_fa.Sigma
    model_iv.ivectors = training_xv
    model_iv.scores = None
    model_iv.nb_thread = 1

    model = dict()
    model["model_iv"] = model_iv
    model["global_diar"] = train_diar
    pickle.dump(model, open(f"{model_cfg['tmp_dir']}/model/model_allies_baseline_{model_cfg['model']['vectors']['type']}v.p", "wb" ))

    return model, model_cfg


def extract_vectors(model, model_cfg, current_diar, root_folder):
    """

    """
    if model_cfg["model"]["type"] == "lium_iv":
        # Load features
        fs_acoustic = sidekit.FeaturesServer(feature_filename_structure=model_cfg['tmp_dir'] + "/feat/{}.h5",
                                             **model_cfg["model"]["vectors"]["ivectors"]["feature_server"])

        current_im = current_diar.id_map()
        # Extract i-vectors
        current_vec = model["model_iv"].train_per_segment(fs_acoustic,current_im, normalization=False)

    elif model_cfg["model"]["type"] == "lium_xv":
        # extract x-vectors
        current_im = current_diar.id_map()
        current_im.start = current_im.start * 160 
        current_im.stop = current_im.stop * 160 
        current_vec = sidekit.nnet.xvector.extract_embeddings(idmap_name=current_im,
                                                              model_filename=model_cfg['tmp_dir']+"/model/best_xtractor.pt",
                                                              data_root_name=root_folder + '/wav/',
                                                              device=torch.device("cuda:0"),
                                                              transform_pipeline=model_cfg["model"]["vectors"]["xvectors"]["transforms"],
                                                              num_thread=10)

    return current_vec

def allies_init_seg(model, system_config, show, file_info, filename, root_folder):
    """

    """
    if isinstance(system_config, str):
        with open(system_config, 'r') as fh:
            model_cfg = yaml.load(fh, Loader=yaml.FullLoader)
    else:
        model_cfg = system_config

    if not os.path.isdir(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/"):
        os.makedirs(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/")

    # Perform first seg if it doesn't exist
    if not os.path.isfile(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/{show}.mdtm"):
        # perform first seg
        if model_cfg["first_seg"]["type"] == "bic":

            if not os.path.isfile(f"{model_cfg['tmp_dir']}/feat/{show}.h5"):
                # Load MFCC
                fe = sidekit.FeaturesExtractor(**model_cfg["model"]["feature_extractor"])

                fe.save(show,
                        channel=0,
                        input_audio_filename=filename,
                        output_feature_filename=f"{model_cfg['tmp_dir']}/feat/{show}.h5")

            fs_seg = sidekit.FeaturesServer(feature_filename_structure=model_cfg['tmp_dir'] + "/feat/{}.h5",
                                            **model_cfg["first_seg"]["feature_server"])

            cep, _= fs_seg.load(show)

            if model_cfg["model"]["vad"]["type"] == "none":
                init_diar = s4d.segmentation.init_seg(cep, show)

            elif model_cfg["model"]["vad"]["type"] == "from_file":
                print(f"Load ref segmentation: {model_cfg['model']['vad']['dir']}/{show}.mdtm")
                #init_diar = s4d.Diar.read_seg(f"{model_cfg['model']['vad']['dir']}/{show}.seg")              
                init_diar = s4d.Diar.read_mdtm(f"{model_cfg['model']['vad']['dir']}/{show}.mdtm")
                init_diar.pad(25)
                init_diar.pack(50)

            if model_cfg["first_seg"]["bic_lin"]:
                # Bic_lin is not useful when loading the reference and hurts a lot
                current_diar = s4d.segmentation.segmentation(cep, init_diar)
                current_diar = s4d.segmentation.bic_linear(cep, current_diar, model_cfg['first_seg']['thr_l'], sr=False)
            else:
                current_diar = init_diar

            if model_cfg["first_seg"]["hac_bic"]:

                fs_seg = sidekit.FeaturesServer(feature_filename_structure=model_cfg['tmp_dir'] + "/feat/{}.h5",
                                                **model_cfg["first_seg"]["hac_feature_server"])
                cep, _= fs_seg.load(show)

                hac = s4d.clustering.hac_bic.HAC_BIC(cep, current_diar, model_cfg['first_seg']['thr_h'], sr=False)
                current_diar = hac.perform()

            if model_cfg["first_seg"]["viterbi"]:
                current_diar = s4d.viterbi.viterbi_decoding(cep, current_diar, model_cfg['first_seg']['thr_vit'])

        else:
            #slidding window
            nfo = soundfile.info(filename)
            # Get beginning and end of the file
            start = 0.0
            end = float(nfo.frames) / 16000.

            current_diar = s4d.Diar()
            current_diar.add_attribut(new_attribut='gender', default='U')
            s = 0.0
            e = float(model_cfg["first_seg"]["sliding_window"])
            idx = 0
            while e < end:
                current_diar.append(show=show,
                                    cluster=f"tmp{idx}",
                                    start=int(s * 100),
                                    stop=int((s + e) * 100),
                                    gender='U')

                idx += 1
                s += float(model_cfg["first_seg"]["sliding_shift"])
                e += float(model_cfg["first_seg"]["sliding_window"])

        #s4d.Diar.write_seg(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/{show}.seg", current_diar)
        allies_write_diar(current_diar, f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/{show}.mdtm")

    # load the first seg
    #first_pass_diar = s4d.Diar.read_seg(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/{show}.seg")
    #first_pass_diar = s4d.Diar.read_mdtm(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/{show}.mdtm")
    #current_diar = copy.deepcopy(first_pass_diar)

    if model_cfg["second_seg"] == "true":

        # Extract segment representation (i-vectors or x-vectors)
        if not os.path.isfile(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/{show}_{model_cfg['model']['vectors']['type']}v.h5"):
            first_pass_vec = extract_vectors(model, model_cfg, current_diar, root_folder)
            first_pass_vec.write(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/{show}_{model_cfg['model']['vectors']['type']}v.h5")
        else:
            print(f"Load {model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/{show}_{model_cfg['model']['vectors']['type']}v.h5")
            first_pass_vec = sidekit.StatServer(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/{show}_{model_cfg['model']['vectors']['type']}v.h5")

        # If the cores, diar and vec don't exist we perform the clustering
        if not os.path.isfile(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/second_th{model_cfg['within_show']['th_w']}/{show}_{model_cfg['model']['vectors']['type']}v.h5") or not os.path.isfile(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/second_th{model_cfg['within_show']['th_w']}/{show}_{model_cfg['model']['vectors']['type']}v_scores.h5") or not os.path.isfile(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/second_th{model_cfg['within_show']['th_w']}/{show}.mdtm"):

            first_pass_diar = s4d.Diar.read_mdtm(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/{show}.mdtm")
            current_diar = copy.deepcopy(first_pass_diar)
            current_vec = copy.deepcopy(first_pass_vec)

            # Perform HAC-xv if necessary
            current_diar, current_vec, scores = perform_iv_seg(model["model_iv"].sn_mean,
                                                               model["model_iv"].sn_cov,
                                                               model["model_iv"].plda_mean,
                                                               model["model_iv"].plda_f,
                                                               model["model_iv"].plda_sigma,
                                                               model_cfg['within_show']['th_w'],
                                                               current_diar,
                                                               current_vec,
                                                               model_cfg["within_show"]["hac_method"])
            current_vec.write(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/second_th{model_cfg['within_show']['th_w']}/{show}_{model_cfg['model']['vectors']['type']}v.h5")
            scores.write(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/second_th{model_cfg['within_show']['th_w']}/{show}_{model_cfg['model']['vectors']['type']}v_scores.h5")
            allies_write_diar(current_diar, f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/second_th{model_cfg['within_show']['th_w']}/{show}.mdtm")

        else:
            # if files exist we load them 
            first_pass_diar = s4d.Diar.read_mdtm(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/{show}.mdtm")
            current_diar = s4d.Diar.read_mdtm(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/second_th{model_cfg['within_show']['th_w']}/{show}.mdtm")
            scores = sidekit.Scores(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/second_th{model_cfg['within_show']['th_w']}/{show}_{model_cfg['model']['vectors']['type']}v_scores.h5")
            current_vec = sidekit.StatServer(f"{model_cfg['tmp_dir']}/seg/first_th{model_cfg['first_seg']['thr_h']}/second_th{model_cfg['within_show']['th_w']}/{show}_{model_cfg['model']['vectors']['type']}v.h5")

    else:
        current_vec = None
        first_pass_vec = None
        scores = None

    return current_diar, first_pass_diar, current_vec, first_pass_vec, scores, model_cfg['within_show']['th_w']


def allies_within_show_HAL(current_diar,
                           first_pass_diar,
                           current_vec,
                           first_pass_vec,
                           scores_per_cluster,
                           model_cfg,
                           model,
                           file_info,
                           user,
                           uem,
                           ref):
    """
    At the moment only correct the diarization using Active Learning
    :param current_diar:
    :param current_vec:
    :param scores:
    :param th:
    :param model_cfg:
    :param model:
    :param file_info:
    :param user:
    :param uem:
    :param ref:
    :return:
    """

    th = model_cfg['within_show']['th_w']

    if len(current_diar.unique("cluster")) < 2:
        hal = False
        der_track = None
        new_diar = current_diar
        new_vec = current_vec
    else:
        hal = True

    while hal:

        # If the mode is "active learning"
        if 'process_all_nodes' in model_cfg['within_show']:
            p_a_n=model_cfg['within_show']['process_all_nodes']
        else:
            p_a_n=False
        if 'selection_method' in model_cfg['within_show']:
            seg_met=model_cfg['within_show']['selection_method']
        else:
            seg_met='longest'
        if 'conditional_questioning' in model_cfg['within_show']:
            cond_q=model_cfg['within_show']['conditional_questioning']
        else:
            cond_q=False
        if 'hac_method' in model_cfg['within_show']:
            hac_method=model_cfg['within_show']['hac_method']
        else:
            hac_method='complete'

        if seg_met in ["cluster_center_score", "max", "min"]:
            # Compute scores between all pairs of segments as they are needed for those criteria
            # Normalize i-vectors
            first_pass_vec.spectral_norm_stat1(model["model_iv"].sn_mean[:1], model["model_iv"].sn_cov[:1])
            # Compute scores
            ndx = sidekit.Ndx(models=first_pass_vec.modelset, testsegs=first_pass_vec.modelset)
            scores_per_segment = sidekit.iv_scoring.fast_PLDA_scoring(first_pass_vec, first_pass_vec, ndx,
                                                                      model["model_iv"].plda_mean,
                                                                      model["model_iv"].plda_f,
                                                                      model["model_iv"].plda_sigma,
                                                                      p_known=0.0,
                                                                      scaling_factor=1.0,
                                                                      check_missing=False)
            scores_per_segment.scoremat = 0.5 * (scores_per_segment.scoremat + scores_per_segment.scoremat.transpose())
        else:
            scores_per_segment = None


        hal, new_diar, new_vec, der_track = active_learning_tree_two_options(current_diar,
                                                                            first_pass_diar,
                                                                            current_vec,
                                                                            scores_per_cluster,
                                                                            scores_per_segment,
                                                                            th,
                                                                            user,
                                                                            file_info,
                                                                            uem,
                                                                            ref,
                                                                            process_all_nodes=p_a_n,
                                                                            clustering_method=hac_method,
                                                                            selection_method=seg_met,
                                                                            conditional_questioning=cond_q)

    return new_diar, new_vec, der_track

def allies_cross_show_clustering(show_idx, model, model_cfg, current_diar, current_vec, th_x, adapt=False):
    """

    """
    if show_idx == 0:
        model["previous_vec"] = copy.deepcopy(current_vec)
        model["previous_diar"] = current_diar

    else:
        #previous_diar.segments += within_diar.segments
        # Apply cross-show clustering
        previous_vec, previous_diar, current_diar = cross_show(model=model,
                                                               model_cfg=model_cfg,
                                                               previous_iv=model["previous_vec"],
                                                               previous_diar=model["previous_diar"],
                                                               within_iv=current_vec,
                                                               within_diar=current_diar,
                                                               th_x=th_x,
                                                               incremental=True)

        model["previous_vec"] = previous_vec
        model["previous_diar"] = previous_diar

        if adapt:
            norm_iv = sidekit.StatServer.read('tmp/model/training_xv.h5'.format('ester_repere'))
            model['model_iv'] = adapt_model(model['model_iv'], previous_vec, norm_iv, 0.2)

    return model, current_diar


def allies_write_diar(current_diar, filename):
    """

    """
    cdiar = copy.deepcopy(current_diar)
    for idx, seg in enumerate(cdiar):
        cdiar.segments[idx]['start'] = float(seg['start']) / 100.
        cdiar.segments[idx]['stop'] = float(seg['stop']) / 100.


    cdiar.sort(['show', 'start'])
    with open(filename, 'w', encoding="utf8") as fic:
        for line in s4d.Diar.to_string_seg(cdiar, time_float=True):
            fic.write(line)


